# Local
from . version import __version__
from . common import last_4

class AuthMethod():
    """
    An abstract class for authentication methods.

    We implement our own interface instead of using the upstream library's interface for
    auth methods because it does not natively support all the different use cases needed
    for supporting all of PagerDuty's public APIs, i.e. the ``Bearer`` method, or
    placing the API credential into a parameter in the body of the request.

    :param secret:
        The API credential to be used for authentication.
    """

    def __init__(self, secret):
        self.secret = secret

    @property
    def auth_header(self) -> dict:
        """
        Generates the header that will be used for authenticating with
        the PagerDuty API
        """
        raise NotImplementedError

    @property
    def auth_param(self) -> dict:
        """
        Generates an authentication parameter to go into the body of the request.
        """
        raise NotImplementedError

    @property
    def secret(self):
        """
        Returns the API secret associated with the authentication method.
        """
        return self._secret

    @secret.setter
    def secret(self, secret):
        self._secret = secret

    @property
    def trunc_secret(self) -> str:
        """
        Returns a truncated version of the API credential for display purposes.
        """
        return last_4(self.secret)

class HeaderAuthMethod(AuthMethod):
    """
    Abstract base class for auth methods that authenticate using request headers.

    In this class, ``auth_param`` is defined such that it injects no parameters into the
    body of the request by default, leaving ``auth_header`` un-implemented.
    """

    @property
    def auth_param(self) -> dict:
        return {}

class BodyParameterAuthMethod(AuthMethod):
    """
    Abstract base class for auth methods that authenticate using a body parameter.

    In this class, ``auth_header`` is defined such that it adds no headers to the
    request, but it leaves ``auth_param`` un-implemented to require its implementation.
    """

    @property
    def auth_header(self):
        return {}



import os
import sys

UNKNOWN_VERSION = "5.*.*"

def get_version_via_tomllib():
    import tomllib
    base_project_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    with open(os.path.join(base_project_path, 'pyproject.toml'), 'rb') as f:
        pkg_meta = tomllib.load(f)
        return pkg_meta.get('project', {}).get('version', UNKNOWN_VERSION)

def get_version():
    if sys.version_info.major == 3 and sys.version_info.minor < 8:
        # There is no way to obtain the version from the package metadata because the
        # necessary importlib features have not yet been added. At some point we need to
        # drop support for these versions. We only care about Python major version 3
        # because version 2 is already not supported and Python 4 isn't out yet.
        return UNKNOWN_VERSION
    else:
        try:
            # Use package metadata introspection to get the version:
            from importlib.metadata import version
            return version(__package__)
        except:
            # No package has been built/installed yet, so this is a stopgap to avoid
            # errors in local unit tests and documentation builds:
            if sys.version_info.minor < 11:
                # tomllib was introduced in 3.11
                return UNKNOWN_VERSION
            else:
                # Use tomllib so the correct version number goes into the doc build:
                return get_version_via_tomllib()

__version__ = get_version()



#!/usr/bin/env python3

import argparse
import socket
import sys
from .events_api_v2_client import EventsApiV2Client

HOSTNAME = socket.gethostname()
DEFAULT_SOURCE = f"pagerduty.cli on {HOSTNAME}"

def run(argv):
    parser = argparse.ArgumentParser(
        description='PagerDuty Events API V2 Command Line Interface'
    )
    parser.add_argument('action', choices=['trigger', 'acknowledge', 'resolve'],
        help='Action to perform: trigger, acknowledge, or resolve an incident')
    parser.add_argument('-k', '--routing-key', required=True,
        help='PagerDuty Events API v2 routing key')
    parser.add_argument('-i', '--dedup-key', default=None,
        help='Deduplication key for the incident')
    parser.add_argument('--description', help='Summary/description of the alert')
    parser.add_argument('--source', help='Source of the alert', default=DEFAULT_SOURCE)

    args = parser.parse_args(argv)

    # Create the Events API client
    client = EventsApiV2Client(args.routing_key)

    try:
        # Handle the different actions
        if args.action == 'trigger':
            if not args.description:
                parser.error("--description is required for trigger action")
            dedup_key = client.trigger(args.description, args.source,
                dedup_key=args.dedup_key)
            print(f"Alert triggered successfully. Deduplication key: {dedup_key}")

        elif args.action == 'acknowledge':
            if not args.dedup_key:
                parser.error("-i/--dedup-key is required for acknowledge action")

            dedup_key = client.acknowledge(args.dedup_key)
            print(f"Alert acknowledged successfully. Deduplication key: {dedup_key}")

        elif args.action == 'resolve':
            if not args.dedup_key:
                parser.error("-i/--dedup-key is required for resolve action")

            dedup_key = client.resolve(args.dedup_key)
            print(f"Alert resolved successfully. Deduplication key: {dedup_key}")

    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)

def main():
    run(sys.argv[1:])

if __name__ == '__main__':
    main()



from . version import __version__

from . common import (
    TEXT_LEN_LIMIT,
    TIMEOUT,
    deprecated_kwarg,
    http_error_message,
    last_4,
    normalize_url,
    plural_name,
    requires_success,
    singular_name,
    successful_response,
    truncate_text,
    try_decoding
)

from . api_client import ApiClient

from . events_api_v2_client import EventsApiV2Client

from . oauth_token_client import OAuthTokenClient

from . rest_api_v2_base_client import (
    ITERATION_LIMIT,
    RestApiV2BaseClient,
    auto_json,
    endpoint_matches,
    infer_entity_wrapper,
    is_path_param,
    resource_url,
    unwrap,
    wrapped_entities
)

from . rest_api_v2_client import (
    CANONICAL_PATHS,
    CURSOR_BASED_PAGINATION_PATHS,
    ENTITY_WRAPPER_CONFIG,
    RestApiV2Client,
    canonical_path,
    entity_wrappers
)

from . jira_cloud_integration_api_client import JiraCloudIntegrationApiClient
from . jira_server_integration_api_client import JiraServerIntegrationApiClient
from . ms_teams_integration_api_client import MsTeamsIntegrationApiClient
from . slack_integration_api_client import SlackIntegrationApiClient
from . slack_integration_connections_api_client import SlackIntegrationConnectionsApiClient

from . errors import (
    Error,
    HttpError,
    ServerHttpError,
    UrlError
)

# For backwards compatibility, __all__ currently includes all of the above. This should
# eventually be cleaned up so that it includes only the most-used interfaces, i.e.
# client classes. That could be done as a minor breaking change in a new future major
# version, i.e. it would break "from pagerduty import *" use cases, which would be an
# opportunity to lessen the impact of a future breaking change that ends the practice of
# importing the helper methods and module configuration globals.
__all__ = [
    'ApiClient',
    'CANONICAL_PATHS',
    'CURSOR_BASED_PAGINATION_PATHS',
    'ENTITY_WRAPPER_CONFIG',
    'Error',
    'EventsApiV2Client',
    'HttpError',
    'ITERATION_LIMIT',
    'JiraCloudIntegrationApiClient',
    'JiraServerIntegrationApiClient',
    'MsTeamsIntegrationApiClient',
    'OAuthTokenClient',
    'RestApiV2BaseClient',
    'RestApiV2Client',
    'ServerHttpError',
    'SlackIntegrationApiClient',
    'SlackIntegrationConnectionsApiClient',
    'TEXT_LEN_LIMIT',
    'TIMEOUT',
    'UrlError',
    'auto_json',
    'canonical_path',
    'deprecated_kwarg',
    'endpoint_matches',
    'entity_wrappers',
    'http_error_message',
    'infer_entity_wrapper',
    'is_path_param',
    'last_4',
    'normalize_url',
    'plural_name',
    'requires_success',
    'resource_url',
    'singular_name',
    'successful_response',
    'truncate_text',
    'try_decoding',
    'unwrap',
    'wrapped_entities',
]



from typing import List

from . rest_api_v2_base_client import (
    CanonicalPath,
    RestApiV2BaseClient
)

CANONICAL_PATHS = [
    '/incidents/{incident_id}/meeting'
]

ENTITY_WRAPPER_CONFIG = {
    # The "create a meeting" endpoint follows classic conventions.
    #
    # This dictionary was intentionally created and left empty so that it is fewer steps
    # to support antipatterns if they get added to this API in the future.
}

class MsTeamsIntegrationApiClient(RestApiV2BaseClient):
    """
    Client for the PagerDuty Microsoft Teams Integration API.

    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    This client provides an abstraction layer for the `PagerDuty MS Teams Integration
    API
    <https://developer.pagerduty.com/api-reference/2a7de89c77dc8-pager-duty-ms-teams-integration-api>`_.

    Its documentation indicates that it only supports the original "Token" style
    authentication and does not support OAuth token ("Bearer") authentication. For that
    reason, its constructor does not accept an "auth_type" argument, and it is assumed
    that the provided API key was generated through the PagerDuty web UI.

    For constructor arguments, see :class:`pagerduty.ApiClient`.
    """

    permitted_methods = ('POST', )

    url = "https://api.pagerduty.com/integration-ms-teams"

    def __init__(self, api_key: str, debug: bool = False):
        super(MsTeamsIntegrationApiClient, self).__init__(api_key,
            auth_type='token', debug=debug)
        self.headers.update({
            'Accept': 'application/json',
        })

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG



# Core
from copy import deepcopy
from typing import Iterator, List, Optional, Tuple, Union
from warnings import warn

# PyPI
from requests import Response

# Local
from . api_client import (
    ApiClient,
    normalize_url
)
from . auth_method import (
    AuthMethod,
    HeaderAuthMethod
)

from . common import (
    requires_success,
    singular_name,
    successful_response,
    truncate_text,
    try_decoding,
    last_4
)
from . errors import (
    ServerHttpError,
    UrlError
)

#######################
### CLIENT DEFAULTS ###
#######################
ITERATION_LIMIT = 10000
"""
The maximum position of a result in classic pagination.

The offset plus limit parameter may not exceed this number. This is enforced server-side
and is not something the client may override. Rather, this value is reproduced in the
client to enable short-circuiting pagination, so that the client can avoid the HTTP 400
error that will result if the sum of the limit and offset parameters exceeds it.

See: `Pagination
<https://developer.pagerduty.com/docs/ZG9jOjExMDI5NTU4-pagination>`_.
"""

################################
### REST API V2 URL HANDLING ###
################################

CanonicalPath = str
"""
Canonical path type (alias of ``str``).

Canonical paths are the bold-typed portion of the path of the URL displayed in the API
reference at the top of each reference page for the given API endpoint. They are
interpreted as patterns, i.e. any part of the path enclosed in curly braces (as
determined by :attr:`pagerduty.rest_api_v2_base_client.is_path_param`) is interpreted as
a variable parameter versus a literal substring of the path.
"""

def canonical_path(paths: List[CanonicalPath], base_url: str, url: str) \
        -> CanonicalPath:
    """
    The canonical path from the API documentation corresponding to a URL.

    This method is used to identify and classify URLs according to which particular API
    endpoint within the client's corresponding API it belongs to, in order to account
    for any antipatterns that the endpoint might have.

    For example, in
    `List a user's contact methods
    <https://developer.pagerduty.com/api-reference/50d46c0eb020d-list-a-user-s-contact-methods>`_,
    the canonical path is ``/users/{id}/contact_methods``.

    :param paths:
        A list of paths supported by the API client. One example of this is
        ``pagerduty.rest_api_v2_client.CANONICAL_PATHS``.
    :param base_url:
        The base URL of the API
    :param url:
        A non-normalized URL (a path or full URL)
    :returns:
        The canonical API path corresponding to the URL.
    """
    full_url = normalize_url(base_url, url)
    # Starting with / after hostname before the query string:
    url_path = full_url.replace(base_url.rstrip('/'), '').split('?')[0]
    # Root node (blank) counts so we include it:
    n_nodes = url_path.count('/')
    # First winnow the list down to paths with the same number of nodes:
    patterns = list(filter(
        lambda p: p.count('/') == n_nodes,
        paths
    ))
    # Match against each node, skipping index zero because the root node always
    # matches, and using the adjusted index "j":
    for i, node in enumerate(url_path.split('/')[1:]):
        j = i+1
        patterns = list(filter(
            lambda p: p.split('/')[j] == node or is_path_param(p.split('/')[j]),
            patterns
        ))
        # Don't break early if len(patterns) == 1, but require an exact match...

    if len(patterns) == 0:
        raise UrlError(f"URL {url} does not match any canonical API path " \
            'supported by this client.')
    elif len(patterns) > 1:
        # If there's multiple matches but one matches exactly, return that.
        if url_path in patterns:
            return url_path

        # ...otherwise this is ambiguous.
        raise Exception(f"Ambiguous URL {url} matches more than one " \
            "canonical path pattern: "+', '.join(patterns)+'; this is likely ' \
            'a bug.')
    else:
        return patterns[0]

def endpoint_matches(endpoint_pattern: str, method: str, path: CanonicalPath) -> bool:
    """
    Whether an endpoint (method and canonical path) matches a given pattern.

    This is used by :attr:`pagerduty.rest_api_v2_base_client.entity_wrappers` for
    finding the appropriate entity wrapper configuration entry to use for a given HTTP
    method and API path.

    :param endpoint_pattern:
        The endpoint pattern in the form ``METHOD PATH`` where ``METHOD`` is the
        HTTP method in uppercase or ``*`` to match all methods, and ``PATH`` is
        a canonical API path.
    :param method:
        The HTTP method.
    :param path:
        The canonical API path
    :returns:
        True or False based on whether the pattern matches the endpoint
    """
    return (
        endpoint_pattern.startswith(method.upper()) \
            or endpoint_pattern.startswith('*')
    ) and endpoint_pattern.endswith(f" {path}")

def is_path_param(path_node: str) -> bool:
    """
    Whether a part of a canonical path represents a variable parameter

    :param path_node:
        The node (value between slashes) in the path
    :returns:
        True if the node represents a variable parameter, False if it is a fixed value
    """
    return path_node.startswith('{') and path_node.endswith('}')

###############################
### ENTITY WRAPPING HELPERS ###
###############################
EntityWrapper = str

EntityWrapping = Optional[EntityWrapper]
"""
Descriptive entity wrapping type.

If a string, it indicates that the entity is wrapped in a single property of the body of
the request or response named after the value of that string. If ``None``, it indicates
that entity wrapping is not enabled or should be ignored, i.e. send the user-supplied
request body in the API request or return the response body without any modification.
"""

EntityWrappingSpec = Tuple[EntityWrapping, EntityWrapping]
"""
Descriptive type that specifies how entity wrapping is configured for an API endoint.

The first member indicates the entity wrapping of the request body. The second indicates
the entity wrapping of the response body. The two may differ.
"""

def entity_wrappers(wrapper_config: dict, method: str, path: CanonicalPath) \
        -> EntityWrappingSpec:
    """
    Obtains entity wrapping information for a given endpoint (canonical path and method)

    The information about the API is dependency-injected as the ``wrapper_config``
    parameter. An example of such a dictionary object is
    ``pagerduty.rest_api_v2_client.ENTITY_WRAPPER_CONFIG``.

    When trying to determine the entity wrapper name for a given API endpoint, the
    dictionary ``wrapper_config`` is first checked for keys that apply to a given
    request method and canonical API path based on a matching logic. If no keys are
    found that match, it is assumed that the API endpoint follows classic entity
    wrapping conventions, and the wrapper name can be inferred based on those
    conventions (see :attr:`pagerduty.rest_api_v2_base_client.infer_entity_wrapper`).
    Any new API that does not follow these conventions should therefore be given an
    entry in the ``wrapper_config`` dictionary in order to properly support it for
    entity wrapping.

    Each of the keys of should be a capitalized HTTP method (or ``*`` to match any
    method), followed by a space, followed by a canonical path. Each value is either a
    tuple with request and response body wrappers (if they differ), a string (if they
    are the same for both cases) or ``None`` (if wrapping is disabled and the data is to
    be marshaled or unmarshaled as-is). Values in tuples can also be None to denote that
    either the request or response is unwrapped.

    An endpoint, under the design logic of this client, is said to have entity wrapping
    if the body (request or response) has only one property containing the content
    requested or transmitted, apart from properties used for pagination. If there
    are any secondary content-bearing properties (other than those used for
    pagination), entity wrapping should be disabled to avoid discarding those
    properties from responses or preventing the use of those properties in request
    bodies.

    :param wrapper_config:
        A dictionary in which the entity wrapper antipattern configuration is specified.
    :param method:
        A HTTP method.
    :param path:
        A canonical API path.
    :returns:
        The entity wrapping specification.
    """
    m = method.upper()
    endpoint = "%s %s"%(m, path)
    match = list(filter(
        lambda k: endpoint_matches(k, m, path),
        wrapper_config.keys()
    ))

    if len(match) == 1:
        # Look up entity wrapping info from the global dictionary and validate:
        wrapper = wrapper_config[match[0]]
        invalid_config_error = 'Invalid entity wrapping configuration for ' \
                    f"{endpoint}: {wrapper}; this is most likely a bug."
        if wrapper is not None and type(wrapper) not in (tuple, str):
            # Catch-all for invalid types.
            raise Exception(invalid_config_error)
        elif wrapper is None or type(wrapper) is str:
            # Both request and response have the same wrapping at this endpoint.
            return (wrapper, wrapper)
        elif type(wrapper) is tuple and len(wrapper) == 2:
            # Endpoint may use different wrapping for request and response bodies.
            #
            # Each element must be either str or None. The first element is the request
            # body wrapper and the second is the response body wrapper. If a value is
            # None, that indicates that the request or response value should be encoded
            # and decoded as-is without modifications.
            if False in [w is None or type(w) is str for w in wrapper]:
                # One or both is neither a string nor None, which is invalid:
                raise Exception(invalid_config_error)
            return wrapper
        else:
            # If a tuple but not of length 2, what are we doing here?
            raise Exception(invalid_config_error)
    elif len(match) == 0:
        # Nothing in entity wrapper config matches. In this case it is assumed
        # that the endpoint follows classic API patterns and the wrapper name
        # can be inferred from the URL and request method:
        wrapper = infer_entity_wrapper(method, path)
        return (wrapper, wrapper)
    else:
        matches_str = ', '.join(match)
        raise Exception(f"{endpoint} matches more than one pattern:" + \
            f"{matches_str}; this is most likely a bug.")

def infer_entity_wrapper(method: str, path: CanonicalPath) -> EntityWrapper:
    """
    Infer the entity wrapper name from the endpoint using orthodox patterns.

    This is based on patterns that are broadly applicable but not universal in
    the v2 REST API, where the wrapper name is predictable from the path and
    method. This is the default logic applied to determine the wrapper name
    based on the path if there is no explicit entity wrapping defined for the
    given path i.e. in :attr:`pagerduty.rest_api_v2_client.ENTITY_WRAPPER_CONFIG` for
    :class:`pagerduty.RestApiV2Client`.

    :param method:
        The HTTP method
    :param path:
        A canonical API path i.e. from
        :attr:`pagerduty.rest_api_v2_client.CANONICAL_PATHS`
    """
    m = method.upper()
    path_nodes = path.split('/')
    if is_path_param(path_nodes[-1]):
        # Singular if it's an individual resource's URL for read/update/delete
        # (named similarly to the second to last node, as the last is its ID and
        # the second to last denotes the API resource collection it is part of):
        return singular_name(path_nodes[-2])
    elif m == 'POST':
        # Singular if creating a new resource by POSTing to the index containing
        # similar resources (named simiarly to the last path node):
        return singular_name(path_nodes[-1])
    else:
        # Plural if listing via GET to the index endpoint, or doing a multi-put:
        return path_nodes[-1]

def unwrap(response: Response, wrapper: EntityWrapping) -> Union[dict, list]:
    """
    Unwraps a wrapped entity from a HTTP response.

    :param response:
        The response object.
    :param wrapper:
        The entity wrapper (string), or None to skip unwrapping
    :returns:
        The value associated with the wrapper key in the JSON-decoded body of
        the response, which is expected to be a dictionary (map).
    """
    body = try_decoding(response)
    endpoint = "%s %s"%(response.request.method.upper(), response.request.url)
    if wrapper is not None:
        # There is a wrapped entity to unpack:
        bod_type = type(body)
        error_msg = f"Expected response body from {endpoint} after JSON-" \
            f"decoding to be a dictionary with a key \"{wrapper}\", but "
        if bod_type is dict:
            if wrapper in body:
                return body[wrapper]
            else:
                keys = truncate_text(', '.join(body.keys()))
                raise ServerHttpError(
                    error_msg + f"its keys are: {keys}",
                    response
                )
        else:
            raise ServerHttpError(
                error_msg + f"its type is {bod_type}.",
                response
            )
    else:
        # Wrapping is disabled for responses:
        return body

###########################
### FUNCTION DECORATORS ###
###########################

def auto_json(method: callable) -> callable:
    """
    Decorator to return the full response body object after decoding from JSON.

    Intended for use on functions that take a URL positional argument followed
    by keyword arguments and return a `requests.Response`_ object.

    The new return value is the JSON-decoded response body (``dict`` or ``list``).
    """
    doc = method.__doc__
    def call(self, url, **kw):
        return try_decoding(successful_response(method(self, url, **kw)))
    call.__doc__ = doc
    return call

def resource_url(method: callable) -> callable:
    """
    API call decorator that allows passing a resource dict as the path/URL

    Most resources returned by the API will contain a ``self`` attribute that is
    the URL of the resource itself.

    Using this decorator allows the implementer to pass either a URL/path or
    such a resource dictionary as the ``path`` argument, thus eliminating the
    need to re-construct the resource URL or hold it in a temporary variable.
    """
    doc = method.__doc__
    name = method.__name__
    def call(self, resource, **kw):
        url = resource
        if type(resource) is dict:
            if 'self' in resource: # passing an object
                url = resource['self']
            else:
                # Unsupported APIs for this feature:
                raise UrlError(f"The dict object passed to {name} in place of a URL "
                    "has no 'self' key and cannot be used in place of an API resource "
                    "path/URL.")
        elif type(resource) is not str:
            name = method.__name__
            raise UrlError(f"Value passed to {name} is not a str or dict with "
                "key 'self'")
        return method(self, url, **kw)
    call.__doc__ = doc
    return call

def wrapped_entities(method: callable) -> callable:
    """
    Decorator to automatically wrap request entities and unwrap response entities.

    Used for defining the ``r{method}`` functions, i.e.
    :attr:`pagerduty.RestApiV2BaseClient.rpost`. It makes the methods always return an
    object representing the resource entity in the response (whether wrapped in a
    root-level property or not) rather than the full response body, after JSON-decoding.
    When making a post / put request, and passing the ``json`` keyword argument to
    specify the content to be JSON-encoded as the body, that keyword argument can be
    either the to-be-wrapped content or the full body including the entity wrapper, and
    the ``json`` keyword argument will be normalized to include the wrapper.

    Methods using this decorator will raise a :class:`pagerduty.HttpError` with its
    ``response`` property being being the `requests.Response`_ object in the case of any
    error, so that the implementer can access it by catching the exception, and thus
    design their own custom logic around different types of error responses.

    :param method:
        Method being decorated. Must take one positional argument after ``self`` that is
        the URL/path to the resource, followed by keyword any number of keyword
        arguments, and must return an object of class `requests.Response`_, and be named
        after the HTTP method but with "r" prepended.
    :returns:
        A callable object; the reformed method
    """
    http_method = method.__name__.lstrip('r')
    doc = method.__doc__
    def call(self, url, **kw):
        pass_kw = deepcopy(kw) # Make a copy for modification
        path = self.canonical_path(url)
        endpoint = "%s %s"%(http_method.upper(), path)
        req_w, res_w = self.entity_wrappers(http_method, path)
        # Validate the abbreviated (or full) request payload, and automatically
        # wrap the request entity for the implementer if necessary:
        if req_w is not None and http_method in ('post', 'put') \
                and 'json' in pass_kw and req_w not in pass_kw['json']:
            pass_kw['json'] = {req_w: pass_kw['json']}

        # Make the request:
        r = successful_response(method(self, url, **pass_kw))

        # Unpack the response:
        return unwrap(r, res_w)
    call.__doc__ = doc
    return call

####################
### AUTH METHODS ###
####################

class TokenAuthMethod(HeaderAuthMethod):
    """
    AuthMethod class for the "token" header authentication style.

    This AuthMethod is used primarily in REST API v2 but also is used in some similar
    integration APIs.
    """
    @property
    def auth_header(self) -> dict:
        return {"Authorization": f"Token token={self.secret}"}

class OAuthTokenAuthMethod(HeaderAuthMethod):
    """
    AuthMethod class for OAuth-created authentication tokens ("Bearer")
    """
    @property
    def auth_header(self) -> dict:
        return {"Authorization": f"Bearer {self.secret}"}

####################
### CLIENT CLASS ###
####################

class RestApiV2BaseClient(ApiClient):
    """
    Abstract base class for all API clients that support APIs similar to REST API v2

    This class implements some common features like numeric pagination that also appear
    and are supported to varying degrees outside of REST API v2.

    :param api_key:
        REST API access token to use for HTTP requests
    :param auth_type:
        The type of credential in use. If authenticating with an OAuth access
        token, this must be set to ``oauth2`` or ``bearer``. This will determine the
        format of the ``Authorization`` header that is sent to the API in each request.
    :param debug:
        Sets :attr:`pagerduty.ApiClient.print_debug`. Set to ``True`` to enable verbose
        command line output.
    """

    api_call_counts = None
    """A dict object recording the number of API calls per endpoint"""

    api_time = None
    """A dict object recording the total time of API calls to each endpoint"""

    default_page_size = 100
    """
    This will be the default number of results requested in each page when
    iterating/querying an index (the ``limit`` parameter).
    """

    def __init__(self, api_key: str, auth_type: str = 'token', debug: bool = False):
        self.api_call_counts = {}
        self.api_time = {}
        self.auth_type = auth_type
        auth_method = self._build_auth_method(api_key)
        super(RestApiV2BaseClient, self).__init__(auth_method, debug=debug)

    def _build_auth_method(self, api_key: str) -> AuthMethod:
        """
        Constructs an AuthMethod according to the configured :attr:`auth_type`

        :param api_key:
            The API credential to use for authentication, and to construct the
            ``AuthMethod`` object.
        """
        return self.auth_type_mapping[self.auth_type](api_key)

    @property
    def auth_type(self) -> str:
        """
        Defines the method of API authentication.

        This value determines how the Authorization header will be set. By default this
        is "token", which will result in the format ``Token token=<api_key>``.

        This property was meant to support the backwards-compatible constructor
        interface where the ``auth_type`` keyword argument selects the appropriate
        ``Authorization`` header format (which internally is done through selecting an
        ``AuthMethod``).
        """
        return self._auth_type

    @auth_type.setter
    def auth_type(self, auth_type: str):
        valid_auth_types = list(self.auth_type_mapping.keys())
        if auth_type not in valid_auth_types:
            raise AttributeError(f"auth_type value must be one of: {valid_auth_types}")
        self._auth_type = auth_type

    @property
    def auth_type_mapping(self) -> dict:
        """
        Defines a mapping of valid :attr:`auth_type` values to AuthMethod classes.
        """
        return {
            'token':  TokenAuthMethod,
            'bearer': OAuthTokenAuthMethod,
            'oauth2': OAuthTokenAuthMethod
        }

    def after_set_api_key(self):
        """
        (Deprecated) Setter hook for setting or updating the authentication method.

        Will be replaced by after_set_auth_method.
        """
        pass

    @property
    def api_key(self) -> str:
        """
        (Deprecated) Property representing the API key used for authentication.

        If this property is updated to swap out the API credential in a preexisting
        client object, the :attr:`auth_type` property MUST be updated first, or the
        authentication header format may end up being incorrect for the type of the new
        credential. This behavior is the same as it has always been since the API client
        first supported Bearer tokens for authentication.

        The setter in previous versions would update the client's ``headers`` property
        using the :attr:`auth_header` property, which depended on the value of
        :attr:`auth_type` at that point in the execution flow. As of version 5.0.0,
        :attr:`auth_type` selects the class when constructing the new
        :attr:`pagerduty.ApiClient.auth_method`.

        Moving forward, the preferred method of changing credentials of a preexisting
        client object will be to set the :attr:`pagerduty.ApiClient.auth_method`
        property directly.
        """
        warn("The api_key property is deprecated; use auth_method instead.")
        return self.auth_method.secret

    @api_key.setter
    def api_key(self, api_key: str):
        """
        (Deprecated) Setter for the API key used for authentication.
        """
        warn("The api_key property is deprecated; use auth_method instead.")
        self.auth_method = self._build_auth_method(api_key)
        self.after_set_api_key()

    def canonical_path(self, url: str) -> CanonicalPath:
        """
        Return the canonical path of a URL for a particular API implementation.

        See: :attr:`pagerduty.rest_api_v2_base_client.canonical_path`

        :param url:
            The URL. Must be supported by the API.
        :returns:
            The canonical path corresponding to the URL.
        """
        return canonical_path(self.canonical_paths, self.url, url)

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        """
        List of canonical paths supported by the particular API client.

        Child classes that do not implement this method do not a-priori support any API
        endpoints for features that require entity wrapping, e.g. pagination.

        This value is used as the first argument to
        :attr:`pagerduty.rest_api_v2_base_client.canonical_path` from
        :attr:`pagerduty.RestApiV2BaseClient.canonical_path`.
        """
        return []

    @property
    def cursor_based_pagination_paths(self) -> List[CanonicalPath]:
        """
        List of paths known by the client to support standard cursor-based pagination.
        """
        return []

    def dict_all(self, path: str, by: str = 'id', **kw) -> dict:
        """
        Dictionary representation of all results from an index endpoint.

        With the exception of ``by``, all keyword arguments passed to this method are
        also passed to :attr:`iter_all`; see the documentation on that method for
        further details.

        :param path:
            The index endpoint URL to use.
        :param by:
            The attribute of each object to use for the key values of the dictionary.
            This is ``id`` by default. Please note, there is no uniqueness validation,
            so if you use an attribute that is not distinct for the data set, this
            function will omit some data in the results. If a property is named that
            the schema of the API requested does not have, this method will raise
            ``KeyError``.
        :param kw:
            Keyword arguments to pass to :attr:`iter_all`.
        :returns:
            A dictionary keyed by the values of the property of each result specified by
            the ``by`` parameter.
        """
        iterator = self.iter_all(path, **kw)
        return {obj[by]:obj for obj in iterator}

    @property
    def entity_wrapper_config(self) -> dict:
        """
        Entity wrapping antipattern specification for the given client.

        This dictionary object is sent to
        :attr:`pagerduty.rest_api_v2_base_client.entity_wrappers` when looking up how
        any given API endpoint wraps (or doesn't wrap) response and request entities;
        refer to the documentation on that method for further details.

        Child classes should implement this method and return appropriate configuration
        to cover all schema antipatterns in the APIs that they support. It is otherwise
        assumed that all endpoints in its corresponding API follow orthodox entity
        wrapping conventions, in which case the wrapper information can be inferred from
        the path itself.
        """
        return {}

    def entity_wrappers(self, http_method: str, path: CanonicalPath) \
            -> EntityWrappingSpec:
        """
        Get the entity-wrapper specification for any given API / API endpoint.

        See: :attr:`pagerduty.rest_api_v2_base_client.entity_wrappers`.

        :param http_method:
            The method of the request.
        :param path:
            The canonical API path of the request.
        :returns:
            The entity wrapper tuple to use in the given request.
        """
        return entity_wrappers(self.entity_wrapper_config, http_method, path)

    def get_total(self, url: str, params: Optional[dict] = None) -> int:
        """
        Gets the total count of records from a classic pagination index endpoint.

        :param url:
            The URL of the API endpoint to query
        :param params:
            An optional dictionary indicating additional parameters to send to the
            endpoint, i.e. filters, time range (``since`` and ``until``), etc. This may
            influence the total, i.e. if specifying a filter that matches a subset of
            possible results.
        :returns:
            The total number of results from the endpoint with the parameters given.
        """
        query_params = deepcopy(params)
        if query_params is None:
            query_params = {}
        query_params.update({
            'total': True,
            'limit': 1,
            'offset': 0
        })
        response = successful_response(self.get(url, params=query_params))
        response_json = try_decoding(response)
        if 'total' not in response_json:
            path = self.canonical_path(url)
            raise ServerHttpError(
                f"Response from endpoint GET {path} lacks a \"total\" property. This " \
                "may be because the endpoint does not support classic pagination, or " \
                "implements it incompletely or incorrectly.",
                response
            )
        return int(response_json['total'])

    def iter_all(self, url, params: Optional[dict] = None,
                page_size: Optional[int] = None, item_hook: Optional[callable] = None,
                total: Optional[bool] = False) -> Iterator[dict]:
        """
        Iterator for the contents of an index endpoint or query.

        Automatically paginates and yields the results in each page, until all
        matching results have been yielded or a HTTP error response is received.

        If the URL to use supports cursor-based pagintation, then this will
        return :attr:`iter_cursor` with the same keyword arguments. Otherwise,
        it implements classic pagination, a.k.a. numeric pagination.

        Each yielded value is a dict object representing a result returned from
        the index. For example, if requesting the ``/users`` endpoint, each
        yielded value will be an entry of the ``users`` array property in the
        response.

        :param url:
            The index endpoint URL to use.
        :param params:
            Additional URL parameters to include.
        :param page_size:
            If set, the ``page_size`` argument will override the
            ``default_page_size`` parameter on the session and set the ``limit``
            parameter to a custom value (default is 100), altering the number of
            pagination results. The actual number of results in the response
            will still take precedence, if it differs; this parameter and
            ``default_page_size`` only dictate what is requested of the API.
        :param item_hook:
            Callable object that will be invoked for each item yielded, i.e. for
            printing progress. It will be called with three parameters: a dict
            representing a given result in the iteration, an int representing the number
            of the item in the series, and a value representing the total number of
            items in the series. If the total isn't knowable, i.e. the ``total``
            parameter is ``False`` or omitted, the value passed in for the third
            argument will be the string value ``"?"``.
        :param total:
            If True, the ``total`` parameter will be included in API calls, and
            the value for the third parameter to the item hook will be the total
            count of records that match the query. Leaving this as False confers
            a small performance advantage, as the API in this case does not have
            to compute the total count of results in the query.
        :yields:
            Results from each page of results.
        """
        # Get entity wrapping and validate that the URL being requested is
        # likely to support pagination:
        path = self.canonical_path(url)
        endpoint = f"GET {path}"

        # Short-circuit to cursor-based pagination if appropriate:
        if path in self.cursor_based_pagination_paths:
            return self.iter_cursor(url, params=params, page_size=page_size,
                item_hook=item_hook)

        nodes = path.split('/')
        if is_path_param(nodes[-1]):
            # NOTE: If this happens for a newer endpoint in REST API v2, and the final
            # path parameter is one of a fixed list of literal strings, the path might
            # need to be added to the EXPAND_PATHS dictionary in
            # scripts/get_path_list/get_path_list.py, after which CANONICAL_PATHS will
            # then need to be updated accordingly based on the new output of the script.
            raise UrlError(f"Path {path} (URL={url}) is formatted like an " \
                "individual resource versus a resource collection. It is " \
                "therefore assumed to not support pagination.")
        _, wrapper = self.entity_wrappers('GET', path)

        if wrapper is None:
            raise UrlError(f"Pagination is not supported for {endpoint}.")

        # Parameters to send:
        data = {
            'limit': (self.default_page_size, page_size)[int(bool(page_size))],
        }
        if total is not None:
            # The reason this is necessary is because urllib.parse.urlencode (called by
            # Requests to serialize querystring parameters) serializes booleans as
            # capitalized values, whereas the PagerDuty API requires lower case
            # "true/false".
            data['total'] = str(total).lower()
        if isinstance(params, (dict, list)):
            # Override defaults with values given:
            data.update(dict(params))

        more = True
        offset = 0
        if params is not None:
            offset = int(params.get('offset', 0))
        n = 0
        while more:
            # Check the offset and limit:
            data['offset'] = offset
            highest_record_index = int(data['offset']) + int(data['limit'])
            if highest_record_index > ITERATION_LIMIT:
                iter_limit = '%d'%ITERATION_LIMIT
                warn(
                    f"Stopping iter_all on {endpoint} at " \
                    f"limit+offset={highest_record_index} " \
                    'as this exceeds the maximum permitted by the API ' \
                    f"({iter_limit}). The set of results may be incomplete."
                )
                return

            # Make the request and validate/unpack the response:
            r = successful_response(
                self.get(url, params=data.copy()),
                context='classic pagination'
            )
            body = try_decoding(r)
            results = unwrap(r, wrapper)

            # Validate and update pagination parameters
            #
            # Note, the number of the results in the actual response is always
            # the most appropriate amount to increment the offset by after
            # receiving each page. If this is the last page, pagination should
            # stop anyways because the ``more`` parameter should evaluate to
            # false.
            #
            # In short, the reasons why we don't trust the echoed ``limit``
            # value or stick to the limit requested and hope the server honors
            # it is that it could potentially result in skipping results or
            # yielding duplicates if there's a mismatch, or potentially issues
            # like PagerDuty/pdpyras#61
            data['limit'] = len(results)
            offset += data['limit']
            more = False
            if 'total' in body:
                total_count = body['total']
            else:
                total_count = '?'
            if 'more' in body:
                more = body['more']
            else:
                warn(
                    f"Response from endpoint GET {path} lacks a \"more\" property and "
                    "therefore does not support pagination. Only results from the "
                    "first request will be yielded. You can use \"rget\" with this "
                    "endpoint instead to avoid this warning."
                )

            # Perform per-page actions on the response data
            for result in results:
                n += 1
                # Call a callable object for each item, i.e. to print progress:
                if hasattr(item_hook, '__call__'):
                    item_hook(result, n, total_count)
                yield result

    def iter_cursor(self, url: str, params: Optional[dict] = None,
                item_hook: Optional[callable] = None,
                page_size: Optional[int] = None) -> Iterator[dict]:
        """
        Iterator for results from an endpoint using cursor-based pagination.

        :param url:
            The index endpoint URL to use.
        :param params:
            Query parameters to include in the request.
        :param item_hook:
            A callable object that accepts 3 positional arguments; see :attr:`iter_all`
            for details on how this argument is used.
        :param page_size:
            Number of results per page of results (the ``limit`` parameter). If
            unspecified, :attr:`default_page_size` will be used.
        :yields:
            Results from each page of results.
        """
        path = self.canonical_path(url)
        if path not in self.cursor_based_pagination_paths:
            raise UrlError(f"{path} does not support cursor-based pagination.")
        _, wrapper = self.entity_wrappers('GET', path)
        user_params = {
            'limit': (self.default_page_size, page_size)[int(bool(page_size))]
        }
        if isinstance(params, (dict, list)):
            # Override defaults with values given:
            user_params.update(dict(params))

        more = True
        next_cursor = None
        total = 0

        while more:
            # Update parameters and request a new page:
            if next_cursor:
                user_params.update({'cursor': next_cursor})
            r = successful_response(
                self.get(url, params=user_params),
                context='cursor-based pagination',
            )

            # Unpack and yield results
            body = try_decoding(r)
            results = unwrap(r, wrapper)
            for result in results:
                total += 1
                if hasattr(item_hook, '__call__'):
                    item_hook(result, total, '?')
                yield result

            # Advance to the next page
            next_cursor = body.get('next_cursor', None)
            more = bool(next_cursor)

    @resource_url
    @auto_json
    def jget(self, url: Union[str, dict], **kw) -> Union[dict, list]:
        """
        Performs a GET request, returning the JSON-decoded body as a dictionary
        """
        return self.get(url, **kw)

    @resource_url
    @auto_json
    def jpost(self, url: Union[str, dict], **kw) -> Union[dict, list]:
        """
        Performs a POST request, returning the JSON-decoded body as a dictionary
        """
        return self.post(url, **kw)

    @resource_url
    @auto_json
    def jput(self, url: Union[str, dict], **kw) -> Optional[Union[dict, list]]:
        """
        Performs a PUT request, returning the JSON-decoded body as a dictionary
        """
        return self.put(url, **kw)

    def list_all(self, url: str, **kw) -> list:
        """
        Returns a list of all objects from a given index endpoint.

        All keyword arguments passed to this function are also passed directly
        to :attr:`iter_all`; see the documentation on that method for details.

        :param url:
            The index endpoint URL to use.
        """
        return list(self.iter_all(url, **kw))

    def postprocess(self, response: Response, suffix: Optional[str] = None):
        """
        Records performance information / request metadata about the API call.

        :param response:
            The `requests.Response`_ object returned by the request method
        :param suffix:
            Optional suffix to append to the key
        :type method: str
        :type response: `requests.Response`_
        :type suffix: str or None
        """
        method = response.request.method.upper()
        url = response.request.url
        status = response.status_code
        request_date = response.headers.get('date', '(missing header)')
        request_id = response.headers.get('x-request-id', '(missing header)')
        request_time = response.elapsed.total_seconds()

        try:
            endpoint = "%s %s"%(method, self.canonical_path(url))
        except UrlError:
            # This is necessary so that profiling can also support using the
            # basic get / post / put / delete methods with APIs that are not yet
            # explicitly supported by inclusion in CANONICAL_PATHS.
            endpoint = "%s %s"%(method, url)
        self.api_call_counts.setdefault(endpoint, 0)
        self.api_time.setdefault(endpoint, 0.0)
        self.api_call_counts[endpoint] += 1
        self.api_time[endpoint] += request_time

        # Request ID / timestamp logging
        self.log.debug("Request completed: #method=%s|#url=%s|#status=%d|"
            "#x_request_id=%s|#date=%s|#wall_time_s=%g", method, url, status,
            request_id, request_date, request_time)
        if int(status/100) == 5:
            self.log.error("PagerDuty API server error (%d)! "
                "For additional diagnostics, contact PagerDuty support "
                "and reference x_request_id=%s / date=%s",
                status, request_id, request_date)

    @resource_url
    @requires_success
    def rdelete(self, resource: Union[str, dict], **kw) -> Response:
        """
        Delete a resource.

        :param resource:
            The path/URL to which to send the request, or a dict object
            representing an API resource that contains an item with key ``self``
            whose value is the URL of the resource.
        :param **kw:
            Custom keyword arguments to pass to ``requests.Session.delete``
        """
        return self.delete(resource, **kw)

    @resource_url
    @wrapped_entities
    def rget(self, resource: Union[str, dict], **kw) -> Union[dict, list]:
        """
        Wrapped-entity-aware GET function.

        Retrieves a resource via GET and returns the wrapped entity in the
        response.

        :param resource:
            The path/URL to which to send the request, or a dict object
            representing an API resource that contains an item with key ``self``
            whose value is the URL of the resource.
        :param **kw:
            Custom keyword arguments to pass to ``requests.Session.get``
        :returns:
            The API response after JSON-decoding and unwrapping
        """
        return self.get(resource, **kw)

    @wrapped_entities
    def rpost(self, path: str, **kw) -> Union[dict, list]:
        """
        Wrapped-entity-aware POST function.

        Creates a resource and returns the created entity if successful.

        :param path:
            The path/URL to which to send the POST request, which should be an
            index endpoint.
        :param **kw:
            Custom keyword arguments to pass to ``requests.Session.post``
        :returns:
            The API response after JSON-decoding and unwrapping
        """
        return self.post(path, **kw)


    @resource_url
    @wrapped_entities
    def rput(self, resource: Union[str, dict], **kw) -> Optional[Union[dict, list]]:
        """
        Wrapped-entity-aware PUT function.

        Update an individual resource, returning the wrapped entity.

        :param resource:
            The path/URL to which to send the request, or a dict object
            representing an API resource that contains an item with key ``self``
            whose value is the URL of the resource.
        :param **kw:
            Custom keyword arguments to pass to ``requests.Session.put``
        :returns:
            The API response after JSON-decoding and unwrapping. In the case of at least
            one Teams endpoint (within REST API v2) and any other future API endpoint
            that responds with 204 No Content, the return value will be None.
        """
        return self.put(resource, **kw)

    @property
    def total_call_count(self) -> int:
        """The total number of API calls made by this instance."""
        return sum(self.api_call_counts.values())

    @property
    def total_call_time(self) -> float:
        """The total time spent making API calls."""
        return sum(self.api_time.values())





from typing import List

from . rest_api_v2_base_client import (
    CanonicalPath,
    RestApiV2BaseClient
)

CANONICAL_PATHS = [
    '/accounts_mappings',
    '/accounts_mappings/{id}'
]

ENTITY_WRAPPER_CONFIG = {
    'GET /accounts_mappings/{id}': None
}

class JiraCloudIntegrationApiClient(RestApiV2BaseClient):
    """
    Client for the PagerDuty Jira Server Integration API.

    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    This client provides an abstraction layer for the `PagerDuty Jira Cloud Integration API
    <https://developer.pagerduty.com/api-reference/70ea43d07719f-pager-duty-jira-cloud-integration-api>`_.

    For constructor arguments, see :class:`pagerduty.RestApiV2BaseClient`.
    """

    permitted_methods = ('GET', )

    url = "https://api.pagerduty.com/integration-jira-cloud"

    def __init__(self, api_key: str, auth_type: str = 'token', debug: bool = False):
        super(JiraCloudIntegrationApiClient, self).__init__(api_key,
            auth_type=auth_type, debug=debug)
        self.headers.update({
            'Accept': 'application/json',
            # All requests in the reference and not just data-bearing create/update
            # methods have this header, so it should also be included in GET:
            'Content-Type': 'application/json'
        })

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG



# Core
from datetime import (
    datetime,
    timedelta,
    timezone
)
from typing import (
    List,
    Optional,
    Tuple,
    Union
)
from warnings import warn
from json.decoder import JSONDecodeError

# PyPI
from requests import Response

# Local
from . errors import (
    Error,
    HttpError,
    ServerHttpError,
    UrlError
)

########################
### DEFAULT SETTINGS ###
########################

DATETIME_FMT = "%Y-%m-%dT%H:%M:%S%z"
"""
The full ISO8601 format used for parsing and formatting datestamps.
"""

TEXT_LEN_LIMIT = 100
"""
The longest permissible length of API content to include in error messages.
"""

TIMEOUT = 60
"""
The default timeout in seconds for any given HTTP request.

Modifying this value will not affect any preexisting API session instances.
Rather, it will only affect new instances. It is recommended to use
:attr:`pagerduty.ApiClient.timeout` to configure the timeout for a given session.
"""


########################
### HELPER FUNCTIONS ###
########################

def datetime_intervals(since: datetime, until: datetime, n=10) \
        -> List[Tuple[datetime, datetime]]:
    """
    Break up a given time interval into a series of smaller consecutive time intervals.

    :param since:
        A datetime object repesenting the beginning of the time interval.
    :param until:
        A datetime object representing the end of the time interval.
    :param n:
        The target number of sub-intervals to generate.
    :returns:
        A list of tuples representing beginnings and ends of sub-intervals within the
        time interval. If the resulting intervals would be less than one second, they
        will be one second.
    """
    total_s = int((until - since).total_seconds())
    if total_s <= 0:
        raise ValueError('Argument "since" must be before "until".')
    elif total_s < n:
        # One-second intervals:
        interval_len = 1
        n_intervals = total_s
    else:
        interval_len = max(int(total_s/n), 1)
        n_intervals = n
    interval_start = since
    intervals = []
    for i in range(n_intervals-1):
        interval_end = interval_start + timedelta(seconds=interval_len)
        intervals.append((interval_start, interval_end))
        interval_start = interval_end
    intervals.append((interval_start, until))
    return intervals

def datetime_to_relative_seconds(datestr: str):
    """
    Convert an ISO8601 string to a relative number of seconds from the current time.
    """
    deadline = strptime(datestr)
    now = datetime.now(timezone.utc)
    return (deadline-now).total_seconds()

def deprecated_kwarg(deprecated_name: str, details: Optional[str] = None,
            method: Optional[str] = None):
    """
    Raises a warning if a deprecated keyword argument is used.

    :param deprecated_name: The name of the deprecated function
    :param details: An optional message to append to the deprecation message
    :param method: An optional method name
    """
    details_msg = ''
    method_msg = ''
    if method is not None:
        method_msg = f" of {method}"
    if details is not None:
        details_msg = f" {details}"
    warn(
        f"Keyword argument \"{deprecated_name}\"{method_msg} is deprecated."+details_msg
    )

def http_error_message(r: Response, context: Optional[str] = None) -> str:
    """
    Formats a message describing a HTTP error.

    :param r:
        The response object.
    :param context:
        A description of when the error was received, or None to not include it
    :returns:
        The message to include in the HTTP error
    """
    received_http_response = bool(r.status_code)
    endpoint = "%s %s"%(r.request.method.upper(), r.request.url)
    context_msg = ""
    if type(context) is str:
        context_msg=f" in {context}"
    if received_http_response and not r.ok:
        err_type = 'unknown'
        if r.status_code / 100 == 4:
            err_type = 'client'
        elif r.status_code / 100 == 5:
            err_type = 'server'
        tr_bod = truncate_text(r.text)
        return f"{endpoint}: API responded with {err_type} error (status " \
            f"{r.status_code}){context_msg}: {tr_bod}"
    elif not received_http_response:
        return f"{endpoint}: Network or other unknown error{context_msg}"
    else:
        return f"{endpoint}: Success (status {r.status_code}) but an " \
            f"expectation still failed{context_msg}"

def last_4(secret: str) -> str:
    """
    Truncate a sensitive value to its last 4 characters

    :param secret: text to truncate
    :returns:
        The truncated text
    """
    return '*'+str(secret)[-4:]

def normalize_url(base_url: str, url: str) -> str:
    """
    Normalize a URL or path to be a complete API URL before query parameters.

    The ``url`` argument may be a path relative to the base URL or a full URL.

    :param url:
        The URL or path to normalize to a full URL.
    :param base_url:
        The base API URL, excluding any trailing slash, i.e.
        "https://api.pagerduty.com"
    :returns:
        The full API URL.
    """
    if url.startswith(base_url):
        return url
    elif not (url.startswith('http://') or url.startswith('https://')):
        return base_url.rstrip('/') + "/" + url.lstrip('/')
    else:
        raise UrlError(
            f"URL {url} does not start with the API base URL {base_url}."
        )

def plural_name(obj_type: str) -> str:
    """
    Pluralizes a name, i.e. the API name from the ``type`` property

    :param obj_type:
        The object type, i.e. ``user`` or ``user_reference``
    :returns:
        The name of the resource, i.e. the last part of the URL for the
        resource's index URL
    """
    if obj_type.endswith('_reference'):
        # Strip down to basic type if it's a reference
        obj_type = obj_type[:obj_type.index('_reference')]
    if obj_type.endswith('y'):
        # Because English
        return obj_type[:-1]+'ies'
    else:
        return obj_type+'s'


def relative_seconds_to_datetime(seconds_remaining: int) -> str:
    """
    Convert a number of seconds in the future to an absolute UTC ISO8601 time string.
    """
    now = datetime.now(timezone.utc)
    target_time = now + timedelta(seconds=seconds_remaining)
    return strftime(target_time)

def requires_success(method: callable) -> callable:
    """
    Decorator that validates HTTP responses.

    Uses :attr:`pagerduty.common.successful_response` for said validation.
    """
    doc = method.__doc__
    def call(self, url, **kw):
        return successful_response(method(self, url, **kw))
    call.__doc__ = doc
    return call

def singular_name(r_name: str) -> str:
    """
    Singularizes a name, i.e. for the entity wrapper in a POST request

    :para r_name:
        The "resource" name, i.e. "escalation_policies", a plural noun that
        forms the part of the canonical path identifying what kind of resource
        lives in the collection there, for an API that follows classic wrapped
        entity naming patterns.
    :returns:
        The singularized name
    """
    if r_name.endswith('ies'):
        # Because English
        return r_name[:-3]+'y'
    else:
        return r_name.rstrip('s')

def strftime(time_obj: datetime) -> str:
    """
    Format a ``datetime`` object to a string

    :param date:
        The ``datetime`` object
    :returns:
        The formatted string
    """
    return time_obj.strftime(DATETIME_FMT)

def strptime(datestr: str) -> datetime:
    """
    Parse a string in full ISO8601 format into a ``datetime.datetime`` object.

    :param datestr:
        Full ISO8601 string representation of the date and time, including time zone
    :returns:
        The datetime object representing the string
    """
    return datetime.strptime(datestr, DATETIME_FMT)

def successful_response(r: Response, context: Optional[str] = None) -> Response:
    """Validates the response as successful.

    Returns the response if it was successful; otherwise, raises
    :attr:`pagerduty.errors.Error`

    :param r:
        Response object corresponding to the response received.
    :param context:
        A description of when the HTTP request is happening, for error reporting
    :returns:
        The response object, if it was successful
    """
    if r.ok and bool(r.status_code):
        return r
    elif r.status_code / 100 == 5:
        raise ServerHttpError(http_error_message(r, context=context), r)
    elif bool(r.status_code):
        raise HttpError(http_error_message(r, context=context), r)
    else:
        raise Error(http_error_message(r, context=context))

def truncate_text(text: str) -> str:
    """Truncates a string longer than :attr:`pagerduty.common.TEXT_LEN_LIMIT`

    :param text: The string to truncate if longer than the limit.
    """
    if len(text) > TEXT_LEN_LIMIT:
        return text[:TEXT_LEN_LIMIT-1]+'...'
    else:
        return text

def try_decoding(r: Response) -> Optional[Union[dict, list, str]]:
    """
    JSON-decode a response body

    Returns the decoded body if successful; raises :class:`pagerduty.ServerHttpError`
    otherwise.

    :param r:
        The response object
    """
    try:
        return r.json()
    except (JSONDecodeError, ValueError) as e:
        if r.text.strip() == '':
            # Some endpoints return HTTP 204 for request types other than delete
            return None
        else:
            raise ServerHttpError(
                "API responded with invalid JSON: " + truncate_text(r.text),
                r,
            )



import urllib.parse
import base64
import hashlib
import secrets
from typing import Optional, Tuple

from copy import deepcopy

from requests import Response

from . api_client import ApiClient
from . auth_method import BodyParameterAuthMethod
from . common import (
    datetime_to_relative_seconds,
    relative_seconds_to_datetime,
    successful_response,
    try_decoding,
    last_4
)
from . errors import ServerHttpError
from . rest_api_v2_client import RestApiV2Client

class ClientCredentialsAuthMethod(BodyParameterAuthMethod):
    """
    AuthMethod used in OAuth token exchange requests.

    This AuthMethod requires the client secret and client ID, which are then transmitted
    to identity.pagerduty.com as parameters in the HTML form-encoded body.

    :param client_secret:
        The client secret used for creating OAuth tokens for the application
    :param client_id:
        The client ID of the provisioned OAuth application
    """

    def __init__(self, client_secret, client_id):
        self.secret = client_secret
        self.client_id = client_id # Additional required parameter

    @property
    def auth_param(self) -> dict:
        return {
            "client_id": self.client_id,
            "client_secret": self.secret
        }

class OAuthTokenClient(ApiClient):
    """
    Client with helpers for performing an OAuth exchange to obtain an access token.

    Tokens obtained using the client can then be used to authenticate REST API clients,
    e.g.

    .. code-block:: python

        oauth_exchange_client = pagerduty.OAuthTokenClient(client_secret, client_id)
        oauth_response = oauth_exchange_client.get_scoped_app_token('read')
        client = pagerduty.RestApiV2Client(oauth_response['access_token'], auth_type='bearer')

    Requires `registering a PagerDuty App
    <https://developer.pagerduty.com/docs/register-an-app>`_ to obtain the necessary
    credentials, and must be used in the context of an OAuth2 authorization flow.

    For further details, refer to:

    - `OAuth Functionality <https://developer.pagerduty.com/docs/oauth-functionality>`_
    - `User OAuth Token via Code Grant <https://developer.pagerduty.com/docs/user-oauth-token-via-code-grant>`_
    - `Obtaining an App OAuth Token <https://developer.pagerduty.com/docs/app-oauth-token>`_

    :param client_secret:
        The secret associated with the application.
    :param client_id:
        The client ID provided when registering the application.
    :param debug:
        Passed to the parent constructor as the ``debug`` argument. See:
        :class:`ApiClient`
    """

    permitted_methods = ('POST',)

    url = 'https://identity.pagerduty.com'

    early_refresh_buffer = 86400
    """
    Number of seconds before the expiration date to perform a token refresh.

    Used when constructing a :class:`RestApiV2Client`; if the current date/time is
    after the expiration date, or less than this number of seconds before it, a token
    refresh is first performed.

    If the application is expected to use the resulting client object for more than this
    amount of time between each new call to :attr:`refresh_client`, this value can be
    set higher.

    By default, this is 24 hours.
    """

    def __init__(self, client_secret: str, client_id: str, debug=False):
        """
        Create an OAuth token client
        """
        auth_method = ClientCredentialsAuthMethod(client_secret, client_id)

        super(OAuthTokenClient, self).__init__(auth_method, debug=debug)

    def amended_auth_response(self, response: Response) -> dict:
        """
        Amends the auth response map as necessary for other functionality.

        :param auth_response:
            A response from the /oauth/token endpoint as a dictionary.
        :returns:
            The same response from the API in dictionary form with an additional key
            ``expiration_date`` containing the expiration date/time of the token in
            ISO8601 format. This value can then be used in :attr:`refresh_client`.
        """
        response_json = try_decoding(successful_response(response))
        if 'expires_in' not in response_json:
            raise ServerHttpError(
                "Auth response did not include expected key \"expires_in\".",
                response
            )
        response_json.update({
            'expiration_date': relative_seconds_to_datetime(response_json['expires_in'])
        })
        return response_json

    def authorize_url(self, scope: str, redirect_uri: str) -> str:
        """
        The authorize URL in PagerDuty that the end user will visit to authorize the app

        :param scope:
            Scope of the OAuth grant requested
        :param redirect_uri:
            The redirect URI in the application that receives the authorization code
            through client redirect from PagerDuty
        :returns:
            The formatted authorize URL.
        """
        return self.get_authorize_url(
            self.auth_method.client_id,
            scope,
            redirect_uri
        )

    @classmethod
    def get_authorize_url(cls, client_id: str, scope: str, redirect_uri: str) -> str:
        """
        Generate an authorize URL.

        This method can be called directly to circumvent the need to produce a client
        secret that is otherwise required to instantiate an object.

        This is the URL that the user initially visits in the application to authorize
        the application, which will ultimately redirect the user to ``redirect_uri`` but
        with the authorization code.

        :param client_id:
            Client ID of the application
        :param scope:
            Scope of the OAuth grant requested
        :param redirect_uri:
            The redirect URI in the application that receives the authorization code
            through client redirect from PagerDuty
        :returns:
            The formatted authorize URL.
        """
        return cls.url + '/oauth/authorize?'+ urllib.parse.urlencode([
            ('client_id', client_id),
            ('redirect_uri', redirect_uri),
            ('response_type', 'code'),
            ('scope', scope)
        ])

    def get_new_token(self, **kw) -> dict:
        """
        Make a token request.

        There should not be any need to call this method directly. Each of the supported
        types of token exchange requests are implemented in other methods:

        * :attr:`get_new_token_from_code`
        * :attr:`get_refreshed_token`
        * :attr:`get_scoped_app_token`
        * :attr:`get_new_token_from_code_with_pkce`

        :param kw:
            Keyword parameters passed to this method are interpreted as parameters to
            set in the body of the request.
        :returns:
            The JSON response from ``identity.pagerduty.com`` as a dictionary after
            amending via :attr:`amended_auth_response`. It should contain a key
            ``access_token`` with the new token and ``expiration_date`` containing the
            date and time when the token will expire in ISO8601 format.
        """
        return self.amended_auth_response(self.post(
            '/oauth/token',
            data = deepcopy(kw),
            headers = { # Overrides the default which would be `application/json`:
                "Content-Type": "application/x-www-form-urlencoded"
            }
        ))

    def get_new_token_from_code(self, auth_code: str, scope: str, redirect_uri: str) \
            -> dict:
        """
        Exchange an authorization code granted by the user for an access token.

        :param auth_code:
            The authorization code received by the application at the redirect URI
            provided.
        :param scope:
            The scope of the authorization request.
        :redirect_uri:
            The redirect URI that was used in the authorization request.
        :returns:
            The JSON response from ``identity.pagerduty.com``, containing a key
            ``access_token`` with the new token, as a dict
        """
        return self.get_new_token(
            grant_type = 'authorization_code',
            code = auth_code,
            scope = scope,
            redirect_uri = redirect_uri
        )

    def get_refreshed_token(self, refresh_token: str) -> dict:
        """
        Obtain a new access token using a refresh token.

        :param refresh_token:
            The refresh token provided in the response of the original access token,
            i.e. in the dict returned by :attr:`get_new_token`
        :returns:
            The JSON response from ``identity.pagerduty.com``, containing a key
            ``access_token`` with the new token, as a dict
        """
        return self.get_new_token(
            grant_type = 'refresh_token',
            refresh_token = refresh_token
        )

    def get_scoped_app_token(self, scope: str):
        """
        Obtain a scoped app token.

        This can be used to grant a server-side app a scoped non-user application token.

        :param scope:
            The scope of the authorization request.
        :returns:
            The JSON response from ``identity.pagerduty.com``, containing a key
            ``access_token`` with the new token, as a dict
        """
        return self.get_new_token(
            grant_type = 'client_credentials',
            scope = scope
        )

    def generate_s256_pkce_params(self) -> Tuple[str, str]:
        """
        Generate PKCE parameters for OAuth authorization with S256 code challenge method

        Creates a code verifier and corresponding SHA256-hashed code challenge parameter
        for OAuth login with PKCE.

        :returns:
            A tuple containing ``(code_verifier, code_challenge)``. The code verifier
            must be temporarily stored so that it can be used in the final step of
            authorization, i.e. as the ``code_verifier`` argument to
            ``get_new_token_from_code_with_pkce``. The code challenge must be included
            in the query parameters of the initial authorize URL, i.e. set it as the the
            ``code_challenge`` argument to :attr:`get_pkce_authorize_url` to generate
            the URL.
        """
        code_verifier = base64.urlsafe_b64encode(
            secrets.token_bytes(32)
        ).decode('utf-8').rstrip('=')
        code_challenge = base64.urlsafe_b64encode(
            hashlib.sha256(code_verifier.encode('utf-8')).digest()
        ).decode('utf-8').rstrip('=')
        return code_verifier, code_challenge

    def get_pkce_authorize_url(self, scope: str, redirect_uri: str,
                code_challenge: str) -> str:
        """
        Generate an OAuth authorization URL with PKCE parameters ("Leg 1 of 3").

        :param scope:
            Scope of the OAuth grant requested
        :param redirect_uri:
            The redirect URI in the application that receives the authorization code
        :param code_challenge:
            The code challenge generated from the code verifier, i.e. as returned by
            :attr:`generate_s256_pkce_params`.
        :returns:
            The formatted PKCE authorize URL
        """
        params = [
            ('client_id', self.client_id),
            ('redirect_uri', redirect_uri),
            ('response_type', 'code'),
            ('scope', scope),
            ('code_challenge', code_challenge),
            ('code_challenge_method', 'S256')
        ]
        return self.url + '/oauth/authorize?' + urllib.parse.urlencode(params)

    def get_new_token_from_code_with_pkce(self, auth_code: str, scope: str,
                redirect_uri: str, code_verifier: str) -> dict:
        """
        Exchange an authorization code for an access token using PKCE (Leg 3 of 3).

        :param auth_code:
            The authorization code received by the application at the redirect URI
        :param scope:
            The scope of the authorization request
        :param redirect_uri:
            The redirect URI that was used in the authorization request
        :param code_verifier:
            The code verifier used to generate the code challenge in Leg 1, i.e. as
            returned by :attr:`generate_s256_pkce_params`.
        :returns:
            The JSON response from ``identity.pagerduty.com``, containing a key
            ``access_token`` with the new token, as a dict.
        """
        return self.get_new_token(
            grant_type='authorization_code',
            code=auth_code,
            scope=scope,
            redirect_uri=redirect_uri,
            code_verifier=code_verifier
        )

    def refresh_client(self, access_token: str, refresh_token: str,
                expiration_date: str, base_url: str = 'https://api.pagerduty.com', **kw
            ) -> Tuple[RestApiV2Client, Optional[dict]]:
        """
        Instantiate and return a :class:`pagerduty.RestApiV2Client` client object

        Performs a token refresh if the current time is later than
        :attr:`early_refresh_buffer` seconds before the expiration date.

        :param access_token:
            The current REST API access token.
        :param refresh_token:
            The refresh token required to refresh the access token.
        :param expiration_date:
            The expiration date of the access token, formatted as an ISO8601 datetime
            string, including the timezone suffix as a UTC offset. The value contained
            in the ``expiration_date`` key of the amended response dictionary returned
            by any of the "get token" methods should be usable as this parameter.
        :param base_url:
            The value to use for the ``url`` attribute of the API client object.
        :param kw:
            Keyword arguments to pass to the constructor of the API client object.
        :returns:
            A tuple containing a :class:`pagerduty.RestApiV2Client` object as its first
            element and the the amended OAuth response if a refresh was performed (and
            ``None`` otherwise) as its second element.
        """
        auth = None
        current_access_token = access_token
        if datetime_to_relative_seconds(expiration_date) < self.early_refresh_buffer:
            auth = self.get_refreshed_token(refresh_token)
            current_access_token = auth['access_token']
        client_kw = deepcopy(kw)
        client_kw.update({'auth_type': 'bearer'})
        client = RestApiV2Client(current_access_token, **client_kw)
        client.url = base_url
        return client, auth



from typing import List

from . rest_api_v2_base_client import (
    CanonicalPath,
    RestApiV2BaseClient
)


CANONICAL_PATHS = [
    '/rules'
]

ENTITY_WRAPPER_CONFIG = {
    # The /rules endpoints follow classic conventions; the wrapper can be inferred.
    #
    # This dictionary was intentionally created and left empty so that it is fewer steps
    # to support antipatterns if they get added to this API in the future.
}

class JiraServerIntegrationApiClient(RestApiV2BaseClient):
    """
    Client for the PagerDuty Jira Server Integration API.

    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    This client provides an abstraction layer for the `PagerDuty Jira Server Integration
    API
    <https://developer.pagerduty.com/api-reference/47fca619cb161-pager-duty-jira-server-integration-api>`_.

    Its documentation indicates that it only supports "Bearer" (OAuth-based)
    authentication and does not support the original token-style API authentication. For
    that reason, its constructor does not accept an "auth_type" argument, and it is
    assumed that the provided OAuth token was generated using an OAuth flow.

    :param access_token:
        OAuth access token (Bearer token) obtained via an OAuth flow, used for HTTP requests.
    :param jira_signature_token:
        Connected Jira instance signature token. This validates the connection between
        PagerDuty and a specific Jira instance.
    :param debug:
        Sets :attr:`pagerduty.ApiClient.print_debug`. Set to ``True`` to enable verbose
        command line output.
    """

    permitted_methods = ('GET', 'POST', 'PUT', 'DELETE')

    url = "https://app.pagerduty.com/integration-jira-service"

    def __init__(self, access_token: str, jira_signature_token: str, debug: bool = False):
        super(JiraServerIntegrationApiClient, self).__init__(access_token, auth_type='bearer', debug=debug)

        self.jira_signature_token = jira_signature_token
        self.headers.update({
            'Accept': 'application/json',
            'x-pagerduty-jira-signature': self.jira_signature_token
        })

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG



from typing import List

from . rest_api_v2_base_client import (
    CanonicalPath,
    RestApiV2BaseClient
)

CANONICAL_PATHS = [
    "/workspaces/{slack_team_id}/connections",
    "/workspaces/{slack_team_id}/connections/{connection_id}"
]

ENTITY_WRAPPER_CONFIG = {
    # Slack Connections
    "GET /workspaces/{slack_team_id}/connections": "slack_connections",
    "POST /workspaces/{slack_team_id}/connections": "slack_connection",
    "PUT /workspaces/{slack_team_id}/connections/{connection_id}": "slack_connection",
}

class SlackIntegrationConnectionsApiClient(RestApiV2BaseClient):
    """
    Client for the PagerDuty Slack Integration API's "Connections" endpoints.

    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    This client provides an abstraction layer for the
    `PagerDuty Slack Integration API
    <https://developer.pagerduty.com/api-reference/56fee4184eabc-pager-duty-slack-integration-api>`_,
    specifically the "Connections" API endpoints, which use a different hostname in
    the base URL, ``app.pagerduty.com``, as opposed to ``api.pagerduty.com``.

    For constructor arguments, see :class:`pagerduty.RestApiV2BaseClient`.
    """

    permitted_methods = ('GET', 'POST', 'PUT', 'DELETE')

    url = "https://app.pagerduty.com/integration-slack"

    def __init__(self, api_key: str, auth_type: str = 'token', debug: bool = False):
        super(SlackIntegrationConnectionsApiClient, self).__init__(api_key,
            auth_type=auth_type, debug=debug)
        self.headers.update({
            'Accept': 'application/json',
        })

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG



# Core
from copy import deepcopy
from datetime import datetime
from sys import getrecursionlimit
from typing import Iterator, List, Optional
from warnings import warn

# Local
from . common import (
    datetime_intervals,
    strftime
)
from . rest_api_v2_base_client import (
    ITERATION_LIMIT,
    CanonicalPath,
    RestApiV2BaseClient,
    canonical_path as canonical_path_common,
    entity_wrappers as entity_wrappers_common,
    wrapped_entities
)

########################
### DEFAULT SETTINGS ###
########################

RECURSION_LIMIT = getrecursionlimit() // 2

ITER_HIST_RECURSION_WARNING_TEMPLATE = """RestApiV2Client.iter_history cannot continue
bisecting historical time intervals because {reason}, but the total number of results in
the current requested time sub-interval ({since_until}) still exceeds the hard limit for
classic pagination, {iteration_limit}. Results will be incomplete.{suggestion}
""".replace("\n", " ")

# List of canonical REST API paths
#
# Supporting a new API for entity wrapping will require adding its patterns to
# this list. If it doesn't follow standard naming conventions, it will also
# require one or more new entries in ENTITY_WRAPPER_CONFIG.
#
# To generate new definitions for CANONICAL_PATHS and
# CURSOR_BASED_PAGINATION_PATHS based on the API documentation's source code,
# use scripts/get_path_list/get_path_list.py

# BEGIN auto-generated content

CANONICAL_PATHS = [
    '/{entity_type}/{id}/change_tags',
    '/{entity_type}/{id}/tags',
    '/abilities',
    '/abilities/{id}',
    '/addons',
    '/addons/{id}',
    '/alert_grouping_settings',
    '/alert_grouping_settings/{id}',
    '/analytics/metrics/incidents/all',
    '/analytics/metrics/incidents/escalation_policies',
    '/analytics/metrics/incidents/escalation_policies/all',
    '/analytics/metrics/incidents/services',
    '/analytics/metrics/incidents/services/all',
    '/analytics/metrics/incidents/teams',
    '/analytics/metrics/incidents/teams/all',
    '/analytics/metrics/pd_advance_usage/features',
    '/analytics/metrics/responders/all',
    '/analytics/metrics/responders/teams',
    '/analytics/raw/incidents',
    '/analytics/raw/incidents/{id}',
    '/analytics/raw/incidents/{id}/responses',
    '/analytics/raw/responders/{responder_id}/incidents',
    '/audit/records',
    '/automation_actions/actions',
    '/automation_actions/actions/{id}',
    '/automation_actions/actions/{id}/invocations',
    '/automation_actions/actions/{id}/services',
    '/automation_actions/actions/{id}/services/{service_id}',
    '/automation_actions/actions/{id}/teams',
    '/automation_actions/actions/{id}/teams/{team_id}',
    '/automation_actions/invocations',
    '/automation_actions/invocations/{id}',
    '/automation_actions/runners',
    '/automation_actions/runners/{id}',
    '/automation_actions/runners/{id}/teams',
    '/automation_actions/runners/{id}/teams/{team_id}',
    '/business_services',
    '/business_services/{id}',
    '/business_services/{id}/account_subscription',
    '/business_services/{id}/subscribers',
    '/business_services/{id}/supporting_services/impacts',
    '/business_services/{id}/unsubscribe',
    '/business_services/impactors',
    '/business_services/impacts',
    '/business_services/priority_thresholds',
    '/change_events',
    '/change_events/{id}',
    '/escalation_policies',
    '/escalation_policies/{id}',
    '/escalation_policies/{id}/audit/records',
    '/event_orchestrations',
    '/event_orchestrations/{id}',
    '/event_orchestrations/{id}/integrations',
    '/event_orchestrations/{id}/integrations/{integration_id}',
    '/event_orchestrations/{id}/integrations/migration',
    '/event_orchestrations/{id}/global',
    '/event_orchestrations/{id}/router',
    '/event_orchestrations/{id}/unrouted',
    '/event_orchestrations/services/{service_id}',
    '/event_orchestrations/services/{service_id}/active',
    '/event_orchestrations/{id}/cache_variables',
    '/event_orchestrations/{id}/cache_variables/{cache_variable_id}',
    '/event_orchestrations/services/{service_id}/cache_variables',
    '/event_orchestrations/services/{service_id}/cache_variables/{cache_variable_id}',
    '/extension_schemas',
    '/extension_schemas/{id}',
    '/extensions',
    '/extensions/{id}',
    '/extensions/{id}/enable',
    '/incident_workflows',
    '/incident_workflows/{id}',
    '/incident_workflows/{id}/instances',
    '/incident_workflows/actions',
    '/incident_workflows/actions/{id}',
    '/incident_workflows/triggers',
    '/incident_workflows/triggers/{id}',
    '/incident_workflows/triggers/{id}/services',
    '/incident_workflows/triggers/{trigger_id}/services/{service_id}',
    '/incidents',
    '/incidents/{id}',
    '/incidents/{id}/alerts',
    '/incidents/{id}/alerts/{alert_id}',
    '/incidents/{id}/business_services/{business_service_id}/impacts',
    '/incidents/{id}/business_services/impacts',
    '/incidents/{id}/custom_fields/values',
    '/incidents/{id}/log_entries',
    '/incidents/{id}/merge',
    '/incidents/{id}/notes',
    '/incidents/{id}/outlier_incident',
    '/incidents/{id}/past_incidents',
    '/incidents/{id}/related_change_events',
    '/incidents/{id}/related_incidents',
    '/incidents/{id}/responder_requests',
    '/incidents/{id}/snooze',
    '/incidents/{id}/status_updates',
    '/incidents/{id}/status_updates/subscribers',
    '/incidents/{id}/status_updates/unsubscribe',
    '/incidents/count',
    '/incidents/custom_fields',
    '/incidents/custom_fields/{field_id}',
    '/incidents/custom_fields/{field_id}/field_options',
    '/incidents/custom_fields/{field_id}/field_options/{field_option_id}',
    '/license_allocations',
    '/licenses',
    '/log_entries',
    '/log_entries/{id}',
    '/log_entries/{id}/channel',
    '/maintenance_windows',
    '/maintenance_windows/{id}',
    '/notifications',
    '/oauth_delegations',
    '/oauth_delegations/revocation_requests/status',
    '/oncalls',
    '/paused_incident_reports/alerts',
    '/paused_incident_reports/counts',
    '/priorities',
    '/response_plays',
    '/response_plays/{id}',
    '/response_plays/{response_play_id}/run',
    '/rulesets',
    '/rulesets/{id}',
    '/rulesets/{id}/rules',
    '/rulesets/{id}/rules/{rule_id}',
    '/schedules',
    '/schedules/{id}',
    '/schedules/{id}/audit/records',
    '/schedules/{id}/overrides',
    '/schedules/{id}/overrides/{override_id}',
    '/schedules/{id}/users',
    '/schedules/preview',
    '/service_dependencies/associate',
    '/service_dependencies/business_services/{id}',
    '/service_dependencies/disassociate',
    '/service_dependencies/technical_services/{id}',
    '/services',
    '/services/{id}',
    '/services/{id}/audit/records',
    '/services/{id}/change_events',
    '/services/{id}/integrations',
    '/services/{id}/integrations/{integration_id}',
    '/services/{id}/rules',
    '/services/{id}/rules/convert',
    '/services/{id}/rules/{rule_id}',
    '/standards',
    '/standards/{id}',
    '/standards/scores/{resource_type}',
    '/standards/scores/{resource_type}/{id}',
    '/status_dashboards',
    '/status_dashboards/{id}',
    '/status_dashboards/{id}/service_impacts',
    '/status_dashboards/url_slugs/{url_slug}',
    '/status_dashboards/url_slugs/{url_slug}/service_impacts',
    '/status_pages',
    '/status_pages/{id}/impacts',
    '/status_pages/{id}/impacts/{impact_id}',
    '/status_pages/{id}/services',
    '/status_pages/{id}/services/{service_id}',
    '/status_pages/{id}/severities',
    '/status_pages/{id}/severities/{severity_id}',
    '/status_pages/{id}/statuses',
    '/status_pages/{id}/statuses/{status_id}',
    '/status_pages/{id}/posts',
    '/status_pages/{id}/posts/{post_id}',
    '/status_pages/{id}/posts/{post_id}/post_updates',
    '/status_pages/{id}/posts/{post_id}/post_updates/{post_update_id}',
    '/status_pages/{id}/posts/{post_id}/postmortem',
    '/status_pages/{id}/subscriptions',
    '/status_pages/{id}/subscriptions/{subscription_id}',
    '/tags',
    '/tags/{id}',
    '/tags/{id}/users',
    '/tags/{id}/teams',
    '/tags/{id}/escalation_policies',
    '/teams',
    '/teams/{id}',
    '/teams/{id}/audit/records',
    '/teams/{id}/escalation_policies/{escalation_policy_id}',
    '/teams/{id}/members',
    '/teams/{id}/notification_subscriptions',
    '/teams/{id}/notification_subscriptions/unsubscribe',
    '/teams/{id}/users/{user_id}',
    '/templates',
    '/templates/{id}',
    '/templates/{id}/render',
    '/templates/fields',
    '/users',
    '/users/{id}',
    '/users/{id}/audit/records',
    '/users/{id}/contact_methods',
    '/users/{id}/contact_methods/{contact_method_id}',
    '/users/{id}/license',
    '/users/{id}/notification_rules',
    '/users/{id}/notification_rules/{notification_rule_id}',
    '/users/{id}/notification_subscriptions',
    '/users/{id}/notification_subscriptions/unsubscribe',
    '/users/{id}/oncall_handoff_notification_rules',
    '/users/{id}/oncall_handoff_notification_rules/{oncall_handoff_notification_rule_id}',
    '/users/{id}/sessions',
    '/users/{id}/sessions/{type}/{session_id}',
    '/users/{id}/status_update_notification_rules',
    '/users/{id}/status_update_notification_rules/{status_update_notification_rule_id}',
    '/users/me',
    '/vendors',
    '/vendors/{id}',
    '/webhook_subscriptions',
    '/webhook_subscriptions/{id}',
    '/webhook_subscriptions/{id}/enable',
    '/webhook_subscriptions/{id}/ping',
    '/workflows/integrations',
    '/workflows/integrations/{id}',
    '/workflows/integrations/connections',
    '/workflows/integrations/{integration_id}/connections',
    '/workflows/integrations/{integration_id}/connections/{id}',
]

CURSOR_BASED_PAGINATION_PATHS = [
    '/audit/records',
    '/automation_actions/actions',
    '/automation_actions/runners',
    '/escalation_policies/{id}/audit/records',
    '/incident_workflows/actions',
    '/incident_workflows/triggers',
    '/schedules/{id}/audit/records',
    '/services/{id}/audit/records',
    '/teams/{id}/audit/records',
    '/users/{id}/audit/records',
    '/workflows/integrations',
    '/workflows/integrations/connections',
    '/workflows/integrations/{integration_id}/connections',
]

# END auto-generated content

HISTORICAL_RECORD_PATHS = [
    '/audit/records',
    '/change_events',
    '/escalation_policies/{id}/audit/records',
    '/incidents',
    '/log_entries',
    '/oncalls',
    '/schedules/{id}/audit/records',
    '/services/{id}/audit/records',
    '/teams/{id}/audit/records',
    '/users/{id}/audit/records'
]

ENTITY_WRAPPER_CONFIG = {
    # Abilities
    'GET /abilities/{id}': None,
    # Add-ons follows orthodox schema patterns
    # Alert grouping settings follows orthodox schema patterns
    # Analytics
    '* /analytics/metrics/incidents/all': None,
    '* /analytics/metrics/incidents/escalation_policies': None,
    '* /analytics/metrics/incidents/escalation_policies/all': None,
    '* /analytics/metrics/incidents/services': None,
    '* /analytics/metrics/incidents/services/all': None,
    '* /analytics/metrics/incidents/teams': None,
    '* /analytics/metrics/incidents/teams/all': None,
    '* /analytics/metrics/pd_advance_usage/features': None,
    '* /analytics/metrics/responders/all': None,
    '* /analytics/metrics/responders/teams': None,
    '* /analytics/raw/incidents': None,
    '* /analytics/raw/incidents/{id}': None,
    '* /analytics/raw/incidents/{id}/responses': None,

    # Automation Actions
    'POST /automation_actions/actions/{id}/invocations': (None,'invocation'),

    # Paused Incident Reports
    'GET /paused_incident_reports/alerts': 'paused_incident_reporting_counts',
    'GET /paused_incident_reports/counts': 'paused_incident_reporting_counts',

    # Business Services
    '* /business_services/{id}/account_subscription': None,
    'POST /business_services/{id}/subscribers': ('subscribers', 'subscriptions'),
    'POST /business_services/{id}/unsubscribe': ('subscribers', None),
    '* /business_services/priority_thresholds': None,
    'GET /business_services/impacts': 'services',
    'GET /business_services/{id}/supporting_services/impacts': 'services',

    # Change Events
    'POST /change_events': None, # why not just use EventsApiV2Client?
    'GET /incidents/{id}/related_change_events': 'change_events',

    # Event Orchestrations
    '* /event_orchestrations': 'orchestrations',
    '* /event_orchestrations/services/{id}': 'orchestration_path',
    '* /event_orchestrations/services/{id}/active': None,
    '* /event_orchestrations/{id}': 'orchestration',
    '* /event_orchestrations/{id}/global': 'orchestration_path',
    '* /event_orchestrations/{id}/integrations/migration': None,
    '* /event_orchestrations/{id}/router': 'orchestration_path',
    '* /event_orchestrations/{id}/unrouted': 'orchestration_path',

    # Extensions
    'POST /extensions/{id}/enable': (None, 'extension'),

    # Incidents
    'PUT /incidents/{id}/merge': ('source_incidents', 'incident'),
    'POST /incidents/{id}/responder_requests': (None, 'responder_request'),
    'POST /incidents/{id}/snooze': (None, 'incident'),
    'POST /incidents/{id}/status_updates': (None, 'status_update'),
    'POST /incidents/{id}/status_updates/subscribers': ('subscribers', 'subscriptions'),
    'POST /incidents/{id}/status_updates/unsubscribe': ('subscribers', None),
    'GET /incidents/{id}/business_services/impacts': 'services',
    'PUT /incidents/{id}/business_services/{business_service_id}/impacts': None,
    '* /incidents/{id}/custom_fields/values': 'custom_fields',
    'POST /incidents/{id}/responder_requests': None,

    # Incident Custom Fields
    '* /incidents/custom_fields': ('field', 'fields'),
    '* /incidents/custom_fields/{field_id}': 'field',

    # Incident Types
    # TODO: Update after this is GA and no longer early-access (for now we are manually
    # excluding the canonical paths in the update)

    # Incident Workflows
    'POST /incident_workflows/{id}/instances': 'incident_workflow_instance',
    'POST /incident_workflows/triggers/{id}/services': ('service', 'trigger'),

    # Response Plays
    'POST /response_plays/{response_play_id}/run': None, # (deprecated)

    # Schedules
    'POST /schedules/{id}/overrides': ('overrides', None),

    # Service Dependencies
    'POST /service_dependencies/associate': 'relationships',

    # Webhooks
    'POST /webhook_subscriptions/{id}/enable': (None, 'webhook_subscription'),
    'POST /webhook_subscriptions/{id}/ping': None,

    # Status Dashboards
    'GET /status_dashboards/{id}/service_impacts': 'services',
    'GET /status_dashboards/url_slugs/{url_slug}': 'status_dashboard',
    'GET /status_dashboards/url_slugs/{url_slug}/service_impacts': 'services',

    # Status Pages
    # Adheres to orthodox API conventions / fully supported via inference from path

    # Tags
    'POST /{entity_type}/{id}/change_tags': None,

    # Teams
    'PUT /teams/{id}/escalation_policies/{escalation_policy_id}': None,
    'POST /teams/{id}/notification_subscriptions': ('subscribables', 'subscriptions'),
    'POST /teams/{id}/notification_subscriptions/unsubscribe': ('subscribables', None),
    'PUT /teams/{id}/users/{user_id}': None,
    'GET /teams/{id}/notification_subscriptions': 'subscriptions',

    # Templates
    'POST /templates/{id}/render': None,

    # Users
    '* /users/{id}/notification_subscriptions': ('subscribables', 'subscriptions'),
    'POST /users/{id}/notification_subscriptions/unsubscribe': ('subscribables', None),
    'GET /users/{id}/sessions': 'user_sessions',
    'GET /users/{id}/sessions/{type}/{session_id}': 'user_session',
    'GET /users/me': 'user',

    # Workflow Integrations
    # Adheres to orthodox API conventions / fully supported via inference from path

    # OAuth Delegations
    'GET /oauth_delegations/revocation_requests/status': None
}

################################
### REST API V2 URL HANDLING ###
################################

def canonical_path(base_url: str, url: str) -> str:
    """
    Return the REST API v2 canonical path for a given URL.

    This method should eventually be deprecated. For now it is included so that the
    changes to unit testing and the base namespace of the module don't have to change
    dramatically and can still use this wrapper.
    """
    return canonical_path_common(CANONICAL_PATHS, base_url, url)

###############################
### ENTITY WRAPPING HELPERS ###
###############################

def entity_wrappers(method: str, path: str) -> tuple:
    """
    Return the REST API v2 canonical path for a given URL.

    This method should eventually be deprecated. For now it is included so that the
    changes to unit testing and the base namespace of the module don't have to change
    dramatically and can still use this wrapper.
    """
    return entity_wrappers_common(ENTITY_WRAPPER_CONFIG, method, path)

################
# CLIENT CLASS #
################

class RestApiV2Client(RestApiV2BaseClient):
    """
    PagerDuty REST API v2 client class.

    Implements abstractions for the particular features of PagerDuty's REST API v2.
    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    :param api_key:
        REST API access token to use for HTTP requests.
    :param default_from:
        The default email address to use in the ``From`` header when making
        API calls using an account-level API access key.
    :param auth_type:
        The type of credential in use. If authenticating with an OAuth access
        token, this must be set to ``oauth2`` or ``bearer``. This will determine the
        format of the ``Authorization`` header that is sent to the API in each request.
    :param debug:
        Sets :attr:`pagerduty.ApiClient.print_debug`. Set to ``True`` to enable verbose
        command line output.
    """

    default_from = None
    """The default value to use as the ``From`` request header"""

    permitted_methods = ('GET', 'PATCH', 'POST', 'PUT', 'DELETE')

    url = 'https://api.pagerduty.com'
    """Base URL of the REST API"""

    def __init__(self, api_key: str, default_from: Optional[str] = None,
                 auth_type: str = "token", debug: bool = False):

        super(RestApiV2Client, self).__init__(api_key, auth_type, debug=debug)

        self.default_from = default_from
        if default_from is not None:
            self.headers.update({
                'From': default_from
            })

        self.headers.update({
            'Accept': 'application/vnd.pagerduty+json;version=2',
        })

    def account_has_ability(self, ability: str) -> bool:
        """
        Test that the account has an ability.

        :param ability:
            The named ability, i.e. ``teams``.
        :returns:
            True or False based on whether the account has the named ability.
        """
        r = self.get(f"/abilities/{ability}")
        if r.status_code == 204:
            return True
        elif r.status_code == 402:
            return False
        elif r.status_code == 403:
            # Stop. Authorization failed. This is expected to be non-transient. This
            # may be added at a later time to ApiClient, i.e. to add a new default
            # action similar to HTTP 401. It would be a be a breaking change...
            raise HttpError(
                "Received 403 Forbidden response from the API. The identity "
                "associated with the credentials does not have permission to "
                "perform the requested action.", r)
        elif r.status_code == 404:
            raise HttpError(
                f"Invalid or unknown ability \"{ability}\"; API responded with status "
                "404 Not Found.", r)
        return False

    def after_set_auth_method(self):
        self._subdomain = None
        self._api_key_access = None

    @property
    def api_key_access(self) -> str:
        """
        Memoized API key access type getter.

        Will be "user" if the API key is a user-level token (all users should
        have permission to create an API key with the same permissions as they
        have in the PagerDuty web UI).

        If the API key in use is an account-level API token (as only a global
        administrator user can create), this property will be "account".
        """
        if not hasattr(self, '_api_key_access') or self._api_key_access is None:
            response = self.get('/users/me')
            if response.status_code == 400:
                message = try_decoding(response).get('error', '')
                if 'account-level access token' in message:
                    self._api_key_access = 'account'
                else:
                    self._api_key_access = None
                    self.log.error("Failed to obtain API key access level; "
                        "the API did not respond as expected.")
                    self.log.debug("Body = %s", truncate_text(response.text))
            else:
                self._api_key_access = 'user'
        return self._api_key_access

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def cursor_based_pagination_paths(self) -> List[CanonicalPath]:
        return CURSOR_BASED_PAGINATION_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG

    def find(self, resource: str, query: str, attribute: str = 'name',
            params: Optional[dict] = None) -> Optional[dict]:
        """
        Finds an object of a given resource type exactly matching a query.

        Works by querying a given resource index endpoint using the ``query`` parameter.
        To use this function on any given resource, the resource's index must support
        the ``query`` parameter; otherwise, the function may not work as expected. If
        the index ignores the parameter, for instance, this function will take much
        longer to return; results will not be constrained to those matching the query,
        and so every result in the index will be downloaded and compared against the
        query up until a matching result is found or all results have been checked.

        The comparison between the query and matching results is case-insenitive. When
        determining uniqueness, APIs are mostly case-insensitive, and therefore objects
        with similar characters but differing case can't even exist. All results (and
        the search query) are for this reason reduced pre-comparison to a common form
        (all-lowercase strings) so that case doesn't need to match in the query argument
        (which is also interpreted by the API as case-insensitive).

        If said behavior differs for a given API, i.e. the uniqueness constraint on a
        field is case-sensitive, it should still return the correct results because the
        search term sent to the index in the querystring is not lower-cased.

        :param resource:
            The name of the resource endpoint to query, i.e.
            ``escalation_policies``
        :param query:
            The string to query for in the the index.
        :param attribute:
            The property of each result to compare against the query value when
            searching for an exact match. By default it is ``name``, but when
            searching for user by email (for example) it can be set to ``email``
        :param params:
            Optional additional parameters to use when querying.
        :returns:
            The dictionary representation of the result, if found; ``None`` will
            be returned if there is no exact match result.
        """
        query_params = {}
        if params is not None:
            query_params.update(params)
        query_params.update({'query': query})
        simplify = lambda s: str(s).lower()
        search_term = simplify(query)
        equiv = lambda s: simplify(s[attribute]) == search_term
        obj_iter = self.iter_all(resource, params=query_params)
        return next(iter(filter(equiv, obj_iter)), None)

    def iter_alert_grouping_settings(self, service_ids: Optional[list] = None,
                limit: Optional[int] = None) -> Iterator[dict]:
        """
        Iterator for the contents of the "List alert grouping settings" endpoint.

        The API endpoint "GET /alert_grouping_settings" has its own unique method of
        pagination. This method provides an abstraction for it similar to what
        :attr:`iter_all` provides for endpoints that implement classic pagination.

        See:
        `List alert grouping settings <https://developer.pagerduty.com/api-reference/b9fe211cc2748-list-alert-grouping-settings>`_

        :param service_ids:
            A list of specific service IDs to which results will be constrained.
        :param limit:
            The number of results retrieved per page. By default, the value
            :attr:`default_page_size` will be used.
        :yields:
            Results from each page in the ``alert_grouping_settings`` response property.
        """
        more = True
        after = None
        page_size = self.default_page_size
        if limit is not None:
            page_size = limit
        while more:
            params = {'limit': page_size}
            if service_ids is not None:
                params['service_ids[]'] = service_ids
            if after is not None:
                params['after'] = after
            page = self.jget('/alert_grouping_settings', params=params)
            for result in page['alert_grouping_settings']:
                yield result
            after = page.get('after', None)
            more = after is not None

    def iter_analytics_raw_incidents(self, filters: dict, order: str = 'desc',
                order_by: str = 'created_at', limit: Optional[int] = None,
                time_zone: Optional[str] = None) -> Iterator[dict]:
        """
        Iterator for raw analytics data on multiple incidents.

        The API endpoint ``POST /analytics/raw/incidents`` has its own unique method of
        pagination. This method provides an abstraction for it similar to
        :attr:`iter_all`.

        See:
        `Get raw data - multiple incidents <https://developer.pagerduty.com/api-reference/c2d493e995071-get-raw-data-multiple-incidents>`_

        :param filters:
            Dictionary representation of the required ``filters`` parameters.
        :param order:
            The order in which to sort results. Must be ``asc`` or ``desc``.
        :param order_by:
            The attribute of results by which to order results. Must be ``created_at``
            or ``seconds_to_resolve``.
        :param limit:
            The number of results to yield per page before requesting the next page. If
            unspecified, :attr:`default_page_size` will be used. The particular API
            endpoint permits values up to 1000.
        :yields:
            Entries of the ``data`` property in the response body from each page
        """
        page_size = self.default_page_size
        if limit is not None:
            page_size = limit
        more = True
        last = None
        while more:
            body = {
                'filters': filters,
                'order': order,
                'order_by': order_by,
                'limit': page_size
            }
            if time_zone is not None:
                body['time_zone'] = time_zone
            if last is not None:
                body['starting_after'] = last
            page = self.jpost('/analytics/raw/incidents', json=body)
            for result in page['data']:
                yield result
            last = page.get('last', None)
            more = page.get('more', False) and last is not None

    def iter_history(self, url: str, since: datetime, until: datetime,
            recursion_depth: int = 0, **kw) -> Iterator[dict]:
        """
        Yield all historical records from an endpoint in a given time interval.

        This method works around the limitation of classic pagination (see
        :attr:`pagerduty.rest_api_v2_base_client.ITERATION_LIMIT`) by recursively
        bisecting the initially-provided time interval until the total number of results
        in each sub-interval is less than the hard pagination limit.

        :param url:
            Index endpoint (API URL) from which to yield results. In the event that a
            cursor-based pagination endpoint is given, this method calls
            :attr:`iter_cursor` directly, as cursor-based pagination has no such
            limitation.
        :param since:
            The beginning of the time interval. This must be a non-naïve datetime object
            (i.e. it must be timezone-aware), in order to format the ``since`` parameter
            when transmitting it to the API such that it unambiguously takes the time
            zone into account. See: `datetime (Python documentation)
            <https://docs.python.org/3/library/datetime.html>`_
        :param until:
            The end of the time interval. A timezone-aware datetime object must be
            supplied for the same reason as for the ``since`` parameter.
        :param kw:
            Custom keyword arguments to pass to the iteration method. Note, if providing
            ``params`` in order to add query string parameters for filtering, the
            ``since`` and ``until`` keys (if present) will be ignored.
        :yields:
            All results from the resource collection API within the time range specified
            by ``since`` and ``until``.
        """
        path = self.canonical_path(url)
        since_until = {
            'since': strftime(since),
            'until': strftime(until)
        }
        iter_kw = deepcopy(kw)
        if path not in HISTORICAL_RECORD_PATHS:
            # Cannot continue; incompatible endpoint that doesn't accept since/until:
            raise UrlError(f"Method iter_history does not support {path}")
        elif path == '/oncalls':
            # Warn for this specific endpoint but continue:
            warn('iter_history may yield duplicate results when used with /oncalls')
        elif path in self.cursor_based_pagination_paths:
            # Short-circuit to iter_cursor:
            iter_kw.setdefault('params', {})
            iter_kw['params'].update(since_until)
            return self.iter_cursor(url, **iter_kw)
        # Obtain the total number of records for the interval:
        query_params = kw.get('params', {})
        query_params.update(since_until)
        total = self.get_total(url, params=query_params)

        no_results = total == 0
        can_fully_paginate = total <= ITERATION_LIMIT
        min_interval_len = int((until - since).total_seconds()) == 1
        stop_recursion = recursion_depth >= RECURSION_LIMIT
        if no_results:
            # Nothing to be done for this interval
            pass
        elif can_fully_paginate or min_interval_len or stop_recursion:
            # Do not subdivide any further; it is either not necessary or not feasible.
            if not can_fully_paginate:
                # Issue a warning log message
                if stop_recursion:
                    reason = 'the recursion depth limit has been reached'
                    suggestion = ' To avoid this issue, try requesting a smaller ' \
                        'initial time interval.'
                elif min_interval_len:
                    reason = 'the time interval is already the minimum length (1s)'
                    # In practice, this scenario can only happen when PagerDuty ingests
                    # and processes, for a single account, >10k alert history events per
                    # second (to use `/log_entries` as an example). There is
                    # unfortunately nothing more that can be done in this case.
                    suggestion = ''
                self.log.warning(ITER_HIST_RECURSION_WARNING_TEMPLATE.format(
                    reason = reason,
                    since_until = str(since_until),
                    iteration_limit = ITERATION_LIMIT,
                    suggestion = suggestion
                ))
            iter_kw.setdefault('params', {})
            iter_kw['params'].update(since_until)
            for item in self.iter_all(url, **iter_kw):
                yield item
        else:
            # If total exceeds maximum, bisect the time window and recurse:
            iter_kw['recursion_depth'] = recursion_depth + 1
            for (sub_since, sub_until) in datetime_intervals(since, until, n=2):
                for item in self.iter_history(url, sub_since, sub_until, **iter_kw):
                    yield item

    def iter_incident_notes(self, incident_id: Optional[str] = None, **kw) \
            -> Iterator[dict]:
        """
        Iterator for incident notes.

        This is a filtered iterator for log entries of type ``annotate_log_entry``.

        :param incident_id:
            Optionally, request log entries for a specific incident. If included, the
            ``team_ids[]`` query parameter will be removed and ignored.
        :param kw:
            Custom keyword arguments to send to :attr:`iter_all`.
        :yields:
            Incident note log entries as dictionary objects
        """
        my_kw = deepcopy(kw)
        my_kw.setdefault('params', {})
        url = '/log_entries'
        if incident_id is not None:
            url = f"/incidents/{incident_id}/log_entries"
            # The teams filter is irrelevant for a specific incident's log entries, so
            # it must be removed if present:
            for key in ('team_ids', 'team_ids[]'):
                if 'params' in my_kw and key in my_kw['params']:
                    self.log.warn(
                        f"iter_incident_notes: query parameter \"{key}\" will be "
                        "ignored because argument incident_id was specified"
                    )
                    del(my_kw['params'][key])
        return iter(filter(
            lambda ile: ile['type'] == 'annotate_log_entry',
            self.iter_all(url, **my_kw)
        ))

    def normalize_params(self, params: dict) -> dict:
        """
        Modify the user-supplied parameters to ease implementation

        Current behavior:

        * If a parameter's value is of type list, and the parameter name does
          not already end in "[]", then the square brackets are appended to keep
          in line with the requirement that all set filters' parameter names end
          in "[]".

        :returns:
            The query parameters after modification
        """
        updated_params = {}
        for param, value in params.items():
            if type(value) is list and not param.endswith('[]'):
                updated_params[param+'[]'] = value
            else:
                updated_params[param] = value
        return updated_params

    def persist(self, resource: str, attr: str, values: dict, update: bool = False) \
            -> dict:
        """
        Finds or creates and returns a resource with a matching attribute

        Given a resource name, an attribute to use as an idempotency key and a
        set of attribute:value pairs as a dict, create a resource with the
        specified attributes if it doesn't exist already and return the resource
        persisted via the API (whether or not it already existed).

        :param resource:
            The URL to use when creating the new resource or searching for an
            existing one. The underlying AP must support entity wrapping to use
            this method with it.
        :param attr:
            Name of the attribute to use as the idempotency key. For instance,
            "email" when the resource is "users" will not create the user if a
            user with the email address given in ``values`` already exists.
        :param values:
            The content of the resource to be created, if it does not already
            exist. This must contain an item with a key that is the same as the
            ``attr`` argument.
        :param update:
            (New in 4.4.0) If set to True, any existing resource will be updated
            with the values supplied.
        """
        if attr not in values:
            raise ValueError("Argument `values` must contain a key equal "
                "to the `attr` argument (expected idempotency key: '%s')."%attr)
        existing = self.find(resource, values[attr], attribute=attr)
        if existing:
            if update:
                original = {}
                original.update(existing)
                existing.update(values)
                if original != existing:
                    existing = self.rput(existing, json=existing)
            return existing
        else:
            return self.rpost(resource, json=values)

    @wrapped_entities
    def rpatch(self, path: str, **kw) -> dict:
        """
        Wrapped-entity-aware PATCH function.

        Currently the only API endpoint that uses or supports this method is "Update
        Workflow Integration Connection": ``PATCH
        /workflows/integrations/{integration_id}/connections/{id}``

        It cannot use the :attr:`resource_url` decorator because the schema in that case
        has no ``self`` property, and so the URL or path must be supplied.

        :param path:
            The URL to be requested
        :param kw:
            Keyword arguments to send to the request function, i.e. ``params``
        :returns:
            The API response after JSON-decoding and unwrapping
        """
        return self.patch(path, **kw)

    @property
    def subdomain(self) -> str:
        """
        Subdomain of the PagerDuty account of the API access token.
        """
        if not hasattr(self, '_subdomain') or self._subdomain is None:
            try:
                url = self.rget('users', params={'limit':1})[0]['html_url']
                self._subdomain = url.split('/')[2].split('.')[0]
            except Error as e:
                self.log.error("Failed to obtain subdomain; encountered error.")
                self._subdomain = None
                raise e
        return self._subdomain



from typing import List

from . rest_api_v2_base_client import (
    CanonicalPath,
    RestApiV2BaseClient
)

CANONICAL_PATHS = [
    "/incidents/{incident_id}/dedicated_channel"
    "/incidents/{incident_id}/notification_channels"
]

ENTITY_WRAPPER_CONFIG = {
    # Slack Dedicated Channels
    "* /incidents/{incident_id}/dedicated_channel": "channel",

    # Slack Notification Channels
    "GET /incidents/{incident_id}/notification_channels": "channels",
    "POST /incidents/{incident_id}/notification_channels": None,
}

class SlackIntegrationApiClient(RestApiV2BaseClient):
    """
    Client for the PagerDuty Slack Integration API.

    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    This client provides an abstraction layer for all of the endpoints of the
    `PagerDuty Slack Integration API
    <https://developer.pagerduty.com/api-reference/56fee4184eabc-pager-duty-slack-integration-api>`_
    except for the "Slack Connections" features, which are supported by
    :class:`pagerduty.SlackIntegrationConnectionsApiClient`.

    For constructor arguments, see :class:`pagerduty.RestApiV2BaseClient`.
    """

    permitted_methods = ('GET', 'POST', 'PUT', 'DELETE')

    url = "https://api.pagerduty.com/integration-slack"

    def __init__(self, api_key: str, auth_type: str = 'token', debug: bool = False):
        super(SlackIntegrationApiClient, self).__init__(api_key, auth_type=auth_type,
            debug=debug)
        self.headers.update({
            'Accept': 'application/json',
        })

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG

